# coding: utf-8

#  Copyright (c) 2019 | Advancing Technology Systems, LLC
#  See LICENSE for any grants of usage, distribution, or modification

from sqlalchemy import CHAR, Column, DateTime, Float, ForeignKey, Index, Integer, SmallInteger, String, text, and_
from sqlalchemy.dialects.mssql import BIT, MONEY, SMALLDATETIME, TINYINT, UNIQUEIDENTIFIER
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship, foreign


class TableBase(object):
    __table_args__ = {"implicit_returning": False}


Base = declarative_base(cls=TableBase)
metadata = Base.metadata


class SysInfo1(Base):
    __tablename__ = "SysInfo1"

    STO_PK = Column("STO_PK", SmallInteger, primary_key=True, nullable=False, unique=True)
    MINCHG = Column("MINCHG", MONEY, nullable=False, server_default=text("(0)"))
    LOGON = Column("LOGON", BIT, nullable=False, server_default=text("(0)"))
    USEPASS = Column("USEPASS", BIT, nullable=False, server_default=text("(0)"))
    PERIOD = Column("PERIOD", SmallInteger, nullable=False, server_default=text("(0)"))
    PAWNDAYS = Column("PAWNDAYS", SmallInteger, nullable=False, server_default=text("(0)"))
    OPTIONDAYS = Column("OPTIONDAYS", SmallInteger, nullable=False, server_default=text("(0)"))
    GRACEDAYS = Column("GRACEDAYS", SmallInteger, nullable=False, server_default=text("(0)"))
    BUYHOLDDAY = Column("BUYHOLDDAY", SmallInteger, nullable=False, server_default=text("(0)"))
    CNTWEEKEND = Column("CNTWEEKEND", BIT, nullable=False, server_default=text("(0)"))
    POLTICPICK = Column("POLTICPICK", BIT, nullable=False, server_default=text("(0)"))
    POLTICMAIL = Column("POLTICMAIL", BIT, nullable=False, server_default=text("(0)"))
    SAMPAWNPOL = Column("SAMPAWNPOL", BIT, nullable=False, server_default=text("(0)"))
    SAMPAWNPUR = Column("SAMPAWNPUR", BIT, nullable=False, server_default=text("(0)"))
    NUMPOLTIC = Column("NUMPOLTIC", SmallInteger, nullable=False, server_default=text("(0)"))
    REPRNTTICK = Column("REPRNTTICK", BIT, nullable=False, server_default=text("(0)"))
    REPRNTPOL = Column("REPRNTPOL", BIT, nullable=False, server_default=text("(0)"))
    PRINTATF = Column("PRINTATF", BIT, nullable=False, server_default=text("(0)"))
    GUNLOG = Column("GUNLOG", BIT, nullable=False, server_default=text("(0)"))
    FEDGUNNUM = Column(
        "FEDGUNNUM", String(25, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')")
    )
    BACKPAYMNT = Column("BACKPAYMNT", BIT, nullable=False, server_default=text("(0)"))
    EXTENDDATE = Column("EXTENDDATE", BIT, nullable=False, server_default=text("(0)"))
    PRICTAGPAW = Column("PRICTAGPAW", BIT, nullable=False, server_default=text("(0)"))
    PRICTAGPUL = Column("PRICTAGPUL", BIT, nullable=False, server_default=text("(0)"))
    RESPRICEPAWN = Column("RESPRICEPAWN", BIT, nullable=False, server_default=text("(0)"))
    PRNSALEREC = Column("PRNSALEREC", BIT, nullable=False, server_default=text("(0)"))
    COMPNDINT = Column("COMPNDINT", BIT, nullable=False, server_default=text("(0)"))
    EMPSLPRCNT = Column("EMPSLPRCNT", Float(53), nullable=False, server_default=text("(0)"))
    SERIALNUM = Column(
        "SERIALNUM", String(12, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')")
    )
    NAME = Column("NAME", String(35, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')"))
    ADDRESS1 = Column(
        "ADDRESS1", String(30, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')")
    )
    ADDRESS2 = Column(
        "ADDRESS2", String(30, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')")
    )
    CITY = Column("CITY", String(20, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')"))
    STATE = Column("STATE", String(2, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')"))
    ZIP = Column("ZIP", String(10, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')"))
    PHONE = Column("PHONE", String(14, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')"))
    STTAXRATE = Column("STTAXRATE", MONEY, nullable=False, server_default=text("(0)"))
    CTTAXRATE = Column("CTTAXRATE", MONEY, nullable=False, server_default=text("(0)"))
    LCTAXRATE = Column("LCTAXRATE", MONEY, nullable=False, server_default=text("(0)"))
    PARTPAY = Column("PARTPAY", BIT, nullable=False, server_default=text("(0)"))
    PARTEXTDAT = Column("PARTEXTDAT", BIT, nullable=False, server_default=text("(0)"))
    FUTUREPAY = Column("FUTUREPAY", BIT, nullable=False, server_default=text("(0)"))
    LAYAWAYDAY = Column("LAYAWAYDAY", SmallInteger, nullable=False, server_default=text("(0)"))
    LAYAWAYPCT = Column("LAYAWAYPCT", Float(53), nullable=False, server_default=text("(0)"))
    BACKUPDATE = Column("BACKUPDATE", SMALLDATETIME)
    MINAGE = Column("MINAGE", SmallInteger, nullable=False, server_default=text("(0)"))
    LATECHRG = Column("LATECHRG", BIT, nullable=False, server_default=text("(0)"))
    LASTUSE = Column("LASTUSE", SMALLDATETIME)
    POLICENUMS = Column("POLICENUMS", BIT, nullable=False, server_default=text("(0)"))
    MTOU = Column("MTOU", BIT, nullable=False, server_default=text("(0)"))
    V1 = Column("V1", Integer, nullable=False, server_default=text("(0)"))
    V2 = Column("V2", Integer, nullable=False, server_default=text("(0)"))
    V3 = Column("V3", Integer, nullable=False, server_default=text("(0)"))
    V4 = Column("V4", Integer, nullable=False, server_default=text("(0)"))
    V5 = Column("V5", Integer, nullable=False, server_default=text("(0)"))
    V6 = Column("V6", Integer, nullable=False, server_default=text("(0)"))
    NSAOTS = Column("NSAOTS", String(8, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')"))
    CASHDRAWER = Column("CASHDRAWER", BIT, nullable=False, server_default=text("(0)"))
    TAXCOLFEE1 = Column("TAXCOLFEE1", MONEY, nullable=False, server_default=text("(0.0)"))
    TAXCOLAMT = Column("TAXCOLAMT", MONEY, nullable=False, server_default=text("(0.0)"))
    TAXCOLFEE2 = Column("TAXCOLFEE2", MONEY, nullable=False, server_default=text("(0.0)"))
    ENTPAWNDAT = Column("ENTPAWNDAT", BIT, nullable=False, server_default=text("(0)"))
    PASSCLOS = Column(
        "PASSCLOS", String(1, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')")
    )
    MODINVQTY = Column("MODINVQTY", BIT, nullable=False, server_default=text("(0)"))
    PRINLOWER = Column("PRINLOWER", BIT, nullable=False, server_default=text("(0)"))
    LAYLABEL = Column("LAYLABEL", BIT, nullable=False, server_default=text("(0)"))
    LOSTTIKCHG = Column("LOSTTIKCHG", BIT, nullable=False, server_default=text("(0)"))
    LOSTTIKAMT = Column("LOSTTIKAMT", MONEY, nullable=False, server_default=text("(0.00)"))
    BACKDRIVE = Column(
        "BACKDRIVE", String(1, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')")
    )
    CTTAXCOL1 = Column("CTTAXCOL1", Float(53), nullable=False, server_default=text("(0)"))
    CTTAXCOLAT = Column("CTTAXCOLAT", Float(53), nullable=False, server_default=text("(0)"))
    CTTAXCOL2 = Column("CTTAXCOL2", Float(53), nullable=False, server_default=text("(0)"))
    LCTAXCOL1 = Column("LCTAXCOL1", Float(53), nullable=False, server_default=text("(0)"))
    LCTAXCOLAT = Column("LCTAXCOLAT", Float(53), nullable=False, server_default=text("(0)"))
    LCTAXCOL2 = Column("LCTAXCOL2", Float(53), nullable=False, server_default=text("(0)"))
    DIFMONFEE = Column("DIFMONFEE", BIT, nullable=False, server_default=text("(0)"))
    DIFPOLITMS = Column("DIFPOLITMS", BIT, nullable=False, server_default=text("(0)"))
    REPURCHTAX = Column("REPURCHTAX", BIT, nullable=False, server_default=text("(0)"))
    DIFPWPLTIK = Column("DIFPWPLTIK", BIT, nullable=False, server_default=text("(0)"))
    STARTPOL = Column("STARTPOL", Integer, nullable=False, server_default=text("(0)"))
    CANSWNPM = Column("CANSWNPM", CHAR(1, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("(0)"))
    CANSW = Column("CANSW", CHAR(8, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')"))
    PAWNRECPT = Column("PAWNRECPT", BIT, nullable=False, server_default=text("(0)"))
    EXTENDLAY = Column("EXTENDLAY", SmallInteger, nullable=False, server_default=text("(1)"))
    ENCODECOST = Column(
        "ENCODECOST", String(11, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')")
    )
    EXTBAKREST = Column("EXTBAKREST", BIT, nullable=False, server_default=text("(0)"))
    EXTRESTREB = Column("EXTRESTREB", BIT, nullable=False, server_default=text("(0)"))
    StorageFee = Column("StorageFee", MONEY, nullable=False, server_default=text("(0)"))
    BLANKLABEL = Column("BLANKLABEL", BIT, nullable=False, server_default=text("(0)"))
    PAWNOPTION = Column("PAWNOPTION", TINYINT, nullable=False, server_default=text("(1)"))
    NEWCUSTLBL = Column("NEWCUSTLBL", BIT, nullable=False, server_default=text("(0)"))
    NONINVCOST = Column("NONINVCOST", BIT, nullable=False, server_default=text("(0)"))
    BCODEINV = Column("BCODEINV", BIT, nullable=False, server_default=text("(0)"))
    BCODERNG = Column("BCODERNG", BIT, nullable=False, server_default=text("(0)"))
    BCODEPWN = Column("BCODEPWN", BIT, nullable=False, server_default=text("(0)"))
    BCODECST = Column("BCODECST", BIT, nullable=False, server_default=text("(0)"))
    BCODEITM = Column("BCODEITM", BIT, nullable=False, server_default=text("(0)"))
    BCODELTG = Column("BCODELTG", BIT, nullable=False, server_default=text("(0)"))
    BCODELTK = Column("BCODELTK", BIT, nullable=False, server_default=text("(0)"))
    BCODERTG = Column("BCODERTG", BIT, nullable=False, server_default=text("(0)"))
    BCODERTK = Column("BCODERTK", BIT, nullable=False, server_default=text("(0)"))
    BCODERTK = Column("BCODERTK", BIT, nullable=False, server_default=text("(0)"))
    POLICEDEPT = Column(
        "POLICEDEPT", String(50, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')")
    )
    POLICEANUM = Column(
        "POLICEANUM", String(25, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')")
    )
    ASKSALEREC = Column("ASKSALEREC", BIT, nullable=False, server_default=text("(0)"))
    GUNAMT = Column("GUNAMT", MONEY, nullable=False, server_default=text("(0.00)"))
    ALLGUNSAMT = Column("ALLGUNSAMT", BIT, nullable=False, server_default=text("(0)"))
    GLHGUNACCT = Column("GLHGUNACCT", Integer, nullable=False, server_default=text("(0)"))
    GUNINTEXIT = Column("GUNINTEXIT", BIT, nullable=False, server_default=text("(0)"))
    BACKCKEXIT = Column("BACKCKEXIT", BIT, nullable=False, server_default=text("(0)"))
    OVERGUNSAL = Column("OVERGUNSAL", BIT, nullable=False, server_default=text("(0)"))
    THERMALLBL = Column("THERMALLBL", BIT, nullable=False, server_default=text("(0)"))
    ONLYCREDIT = Column("ONLYCREDIT", BIT, nullable=False, server_default=text("(0)"))
    CANADATAXS = Column("CANADATAXS", BIT, nullable=False, server_default=text("(0)"))
    CASHCRDPAS = Column(
        "CASHCRDPAS", String(2, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')")
    )
    GLPRTCHECK = Column("GLPRTCHECK", BIT, nullable=False, server_default=text("(0)"))
    FEDEXEMPT = Column("FEDEXEMPT", Float(53), nullable=False, server_default=text("(0)"))
    IDNUMBERS = Column(
        "IDNUMBERS", String(40, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')")
    )
    PRINTCHECK = Column("PRINTCHECK", BIT, nullable=False, server_default=text("(0)"))
    BENFTEXMPT = Column(
        "BENFTEXMPT", String(2, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')")
    )
    GLBENFTACT = Column("GLBENFTACT", Integer, nullable=False, server_default=text("(0)"))
    GLEICACCT = Column("GLEICACCT", Integer, nullable=False, server_default=text("(0)"))
    PAYCHEKQTR = Column("PAYCHEKQTR", Float(53), nullable=False, server_default=text("(0)"))
    YEARPRINT = Column("YEARPRINT", BIT, nullable=False, server_default=text("(0)"))
    RECONCLPAY = Column("RECONCLPAY", BIT, nullable=False, server_default=text("(0)"))
    PWBYCHECKS = Column(
        "PWBYCHECKS", String(12, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')")
    )
    MEMODAYCHG = Column(
        "MEMODAYCHG", String(4, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')")
    )
    MEMOREBATE = Column(
        "MEMOREBATE", String(18, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')")
    )
    LCNSBACK = Column("LCNSBACK", Float(53), nullable=False, server_default=text("(0)"))
    POLLSTUPTO = Column(
        "POLLSTUPTO", String(24, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')")
    )
    DLSCAN = Column("DLSCAN", String(17, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')"))
    DATACOLECT = Column(
        "DATACOLECT", String(17, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')")
    )
    PROREPSTRT = Column("PROREPSTRT", Integer, nullable=False, server_default=text("(0)"))
    CREDITSALE = Column(
        "CREDITSALE", String(9, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')")
    )
    CRDSALACCT = Column("CRDSALACCT", Integer, nullable=False, server_default=text("(0)"))
    CREDSALMSG = Column(
        "CREDSALMSG", String(80, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')")
    )
    CRDEXPACCT = Column("CRDEXPACCT", Integer, nullable=False, server_default=text("(0)"))
    CRDSVCACCT = Column("CRDSVCACCT", Integer, nullable=False, server_default=text("(0)"))
    CHECKADVNC = Column(
        "CHECKADVNC", String(17, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')")
    )
    TITLEPAWN = Column(
        "TITLEPAWN", String(17, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')")
    )
    CKADVGLACT = Column(
        "CKADVGLACT", String(24, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')")
    )
    TITLPRTDEV = Column(
        "TITLPRTDEV", String(12, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')")
    )
    NOUTAOTS = Column(
        "NOUTAOTS", String(10, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')")
    )
    LATEFEEQST = Column(
        "LATEFEEQST", String(13, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')")
    )
    CustSalRec = Column("CustSalRec", BIT, nullable=False, server_default=text("(0)"))
    OnlyTitle = Column("OnlyTitle", BIT, nullable=False, server_default=text("(0)"))
    Auction = Column(
        "Auction", String(17, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')")
    )
    SY1_id = Column("SY1_id", UNIQUEIDENTIFIER, server_default=text("(newid())"))
    MultiMgmt = Column("MultiMgmt", BIT, nullable=False, server_default=text("(0)"))
    CreditCard = Column("CreditCard", BIT, nullable=False, server_default=text("(0)"))
    OnlineSale = Column("OnlineSale", BIT, nullable=False, server_default=text("(0)"))
    PoliceComm = Column("PoliceComm", BIT, nullable=False, server_default=text("(0)"))
    V7 = Column("V7", Integer, nullable=False, server_default=text("(1)"))
    V8 = Column("V8", Integer, nullable=False, server_default=text("(1)"))
    V9 = Column("V9", Integer, nullable=False, server_default=text("(1)"))
    V10 = Column("V10", Integer, nullable=False, server_default=text("(1)"))
    V11 = Column("V11", Integer, nullable=False, server_default=text("(1)"))
    V12 = Column("V12", Integer, nullable=False, server_default=text("(1)"))
    V13 = Column("V13", Integer, nullable=False, server_default=text("(1)"))
    V14 = Column("V14", Integer, nullable=False, server_default=text("(1)"))
    CustPoints = Column("CustPoints", BIT, nullable=False, server_default=text("(0)"))
    V15 = Column("V15", Integer, nullable=False, server_default=text("(1)"))
    V16 = Column("V16", Integer, nullable=False, server_default=text("(1)"))
    DlScan2 = Column("DlScan2", BIT, nullable=False, server_default=text("(0)"))
    V17 = Column("V17", Integer, nullable=False, server_default=text("(1)"))
    V18 = Column("V18", Integer, nullable=False, server_default=text("(1)"))
    Biometrix = Column("Biometrix", BIT, nullable=False, server_default=text("(0)"))
    V19 = Column("V19", Integer, nullable=False, server_default=text("(1)"))
    V20 = Column("V20", Integer, nullable=False, server_default=text("(1)"))
    ShareDrawers = Column("ShareDrawers", BIT, nullable=False, server_default=text("(0)"))
    DebitLoad = Column("DebitLoad", BIT, nullable=False, server_default=text("(0)"))
    V21 = Column("V21", Integer, nullable=False, server_default=text("(1)"))
    V22 = Column("V22", Integer, nullable=False, server_default=text("(1)"))
    SecureCam = Column("SecureCam", BIT, nullable=False, server_default=text("(0)"))
    V23 = Column("V23", Integer, nullable=False, server_default=text("(1)"))
    V24 = Column("V24", Integer, nullable=False, server_default=text("(1)"))
    Audit = Column("Audit", BIT, nullable=False, server_default=text("(0)"))
    V25 = Column("V25", Integer, nullable=False, server_default=text("(1)"))
    V26 = Column("V26", Integer, nullable=False, server_default=text("(1)"))
    Analysis = Column("Analysis", BIT, nullable=False, server_default=text("(0)"))
    V27 = Column("V27", Integer, nullable=False, server_default=text("(1)"))
    V28 = Column("V28", Integer, nullable=False, server_default=text("(1)"))
    ExcelExprt = Column("ExcelExprt", BIT, nullable=False, server_default=text("(0)"))
    V29 = Column("V29", Integer, nullable=False, server_default=text("(1)"))
    V30 = Column("V30", Integer, nullable=False, server_default=text("(1)"))
    RTimePhys = Column("RTimePhys", BIT, nullable=False, server_default=text("(0)"))
    V31 = Column("V31", Integer, nullable=False, server_default=text("(1)"))
    V32 = Column("V32", Integer, nullable=False, server_default=text("(1)"))
    RFID = Column("RFID", BIT, nullable=False, server_default=text("(0)"))
    V33 = Column("V33", Integer, nullable=False, server_default=text("(1)"))
    V34 = Column("V34", Integer, nullable=False, server_default=text("(1)"))
    LastUpdatedUSR_ID = Column("LastUpdatedUSR_ID", UNIQUEIDENTIFIER)
    BuySell = Column("BuySell", BIT, nullable=False, server_default=text("(0)"))
    V35 = Column("V35", Integer, nullable=False, server_default=text("(1)"))
    V36 = Column("V36", Integer, nullable=False, server_default=text("(1)"))
    BuySellLR = Column("BuySellLR", BIT, nullable=False, server_default=text("(0)"))
    V37 = Column("V37", Integer, nullable=False, server_default=text("(1)"))
    V38 = Column("V38", Integer, nullable=False, server_default=text("(1)"))
    Lite = Column("Lite", BIT, nullable=False, server_default=text("(0)"))
    V39 = Column("V39", Integer, nullable=False, server_default=text("(1)"))
    V40 = Column("V40", Integer, nullable=False, server_default=text("(1)"))
    EmpDrwOptn = Column("EmpDrwOptn", BIT, nullable=False, server_default=text("(0)"))
    V41 = Column("V41", Integer, nullable=False, server_default=text("(1)"))
    V42 = Column("V42", Integer, nullable=False, server_default=text("(1)"))
    GunLogOptn = Column("GunLogOptn", BIT, nullable=False, server_default=text("(0)"))
    V43 = Column("V43", Integer, nullable=False, server_default=text("(1)"))
    V44 = Column("V44", Integer, nullable=False, server_default=text("(1)"))
    MenuLvOptn = Column("MenuLvOptn", BIT, nullable=False, server_default=text("(0)"))
    V45 = Column("V45", Integer, nullable=False, server_default=text("(1)"))
    V46 = Column("V46", Integer, nullable=False, server_default=text("(1)"))
    TextMsg = Column("TextMsg", BIT, nullable=False, server_default=text("(0)"))
    V47 = Column("V47", Integer, nullable=False, server_default=text("(1)"))
    V48 = Column("V48", Integer, nullable=False, server_default=text("(1)"))
    ElecScale = Column("ElecScale", BIT, nullable=False, server_default=text("(0)"))
    V49 = Column("V49", Integer, nullable=False, server_default=text("(1)"))
    V50 = Column("V50", Integer, nullable=False, server_default=text("(1)"))
    PolDown = Column("PolDown", BIT, nullable=False, server_default=text("(0)"))
    V51 = Column("V51", Integer, nullable=False, server_default=text("(1)"))
    V52 = Column("V52", Integer, nullable=False, server_default=text("(1)"))
    QuickQuote = Column("QuickQuote", BIT, nullable=False, server_default=text("(0)"))
    V53 = Column("V53", Integer, nullable=False, server_default=text("(1)"))
    V54 = Column("V54", Integer, nullable=False, server_default=text("(1)"))
    VMNetPro = Column("VMNetPro", BIT, nullable=False, server_default=text("(0)"))
    V55 = Column("V55", Integer, nullable=False, server_default=text("(1)"))
    V56 = Column("V56", Integer, nullable=False, server_default=text("(1)"))
    AdvReports = Column("AdvReports", BIT, nullable=False, server_default=text("(0)"))
    V57 = Column("V57", Integer, nullable=False, server_default=text("(1)"))
    V58 = Column("V58", Integer, nullable=False, server_default=text("(1)"))
    keydx = Column("keydx", BIT, nullable=False, server_default=text("(0)"))
    MultiLang = Column("MultiLang", BIT, nullable=False, server_default=text("(0)"))
    V59 = Column("V59", Integer, nullable=False, server_default=text("(1)"))
    V60 = Column("V60", Integer, nullable=False, server_default=text("(1)"))
    ProtPlan = Column("ProtPlan", BIT, nullable=False, server_default=text("(0)"))
    V61 = Column("V61", Integer, nullable=False, server_default=text("(1)"))
    V62 = Column("V62", Integer, nullable=False, server_default=text("(1)"))
    RmtPay = Column("RmtPay", BIT, nullable=False, server_default=text("(0)"))
    V63 = Column("V63", Integer, nullable=False, server_default=text("(1)"))
    V64 = Column("V64", Integer, nullable=False, server_default=text("(1)"))
    PoleDisp = Column("PoleDisp", BIT, nullable=False, server_default=text("((0))"))
    V65 = Column("V65", Integer, nullable=False, server_default=text("((1))"))
    V66 = Column("V66", Integer, nullable=False, server_default=text("((1))"))
    AcuSport = Column("AcuSport", BIT, nullable=False, server_default=text("((0))"))
    V67 = Column("V67", Integer, nullable=False, server_default=text("((1))"))
    V68 = Column("V68", Integer, nullable=False, server_default=text("((1))"))

    @property
    def name(self):
        return str(self.NAME).strip().title()

    @property
    def address(self):
        return str(
            str(self.ADDRESS1).strip().title()
            + ", "
            + ((str(self.ADDRESS2).strip().title() + ", ") if len(str(self.ADDRESS2).strip()) > 0 else "")
            + str(self.CITY).strip().title()
            + ", "
            + str(self.STATE).strip().upper()
            + " "
            + str(self.ZIP).strip()
        )

    @property
    def phone(self):
        return str(self.PHONE).strip()

    def __repr__(self):
        return self.name


class C_Rates(Base):
    __tablename__ = "c_rates"

    rate_pk = Column(Integer, primary_key=True)
    STO_PK = Column(SmallInteger, ForeignKey("pawn.STORE_NO"), server_default=text("((1))"))
    DESCRIPT = Column(String(20, "SQL_Latin1_General_CP1_CI_AS"), server_default=text("('')"))
    PERIOD = Column(SmallInteger, server_default=text("((0))"))
    AllHigh = Column(BIT, server_default=text("((1))"))
    HD = Column(MONEY, server_default=text("((0))"))
    DP = Column(String(1, "SQL_Latin1_General_CP1_CI_AS"), server_default=text("('')"))
    AP = Column(MONEY, server_default=text("((0))"))
    MINAMT = Column(MONEY, server_default=text("((0))"))
    MAXAMT = Column(MONEY, server_default=text("((0))"))
    ONECHRG = Column(MONEY, server_default=text("((0))"))
    MONCHRG = Column(MONEY, server_default=text("((0))"))
    Interest = Column(MONEY, server_default=text("((0))"))
    NewRateTable = Column(String(20, "SQL_Latin1_General_CP1_CI_AS"), server_default=text("('')"))
    NewDefaults = Column(BIT, server_default=text("((0))"))
    cRates_id = Column(UNIQUEIDENTIFIER, server_default=text("(newid())"))
    SimpleInterest = Column(BIT, server_default=text("((0))"))
    LastUpdatedUSR_ID = Column(UNIQUEIDENTIFIER)

    Index("IDX_c_rates_Sto_PK_Descript", "STO_PK", "DESCRIPT")

    pawn_rate = relationship(
        "Pawn",
        primaryjoin="C_Rates.STO_PK == Pawn.STORE_NO and C_Rates.DESCRIPT == Pawn.RateTable",
        backref="c_rates",
    )

    def __repr__(self):
        return self.DESCRIPT.strip()


class LookupC(Base):
    __tablename__ = "Lookup_C"

    Sto_PK = Column(SmallInteger, nullable=False, server_default=text("(1)"))
    lc_pk = Column(Integer, primary_key=True, server_default=text("(1)"))
    lc_Descript = Column(String(20, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')"))
    LB_FK = Column(Integer, nullable=False, index=True, server_default=text("(0)"))
    LUC_id = Column(UNIQUEIDENTIFIER, unique=True, server_default=text("(newid())"))
    LastUpdatedUSR_ID = Column(UNIQUEIDENTIFIER)


class StatCode(Base):
    __tablename__ = "StatCode"
    __table_args__ = (Index("IDX_StatCode_Trans_Status", "trans", "status"),)

    cType = Column(String(8, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')"))
    trans = Column(
        String(6, "SQL_Latin1_General_CP1_CI_AS"), primary_key=True, nullable=False, server_default=text("('')")
    )
    status = Column(
        String(1, "SQL_Latin1_General_CP1_CI_AS"), primary_key=True, nullable=False, server_default=text("('')")
    )
    DESCRIPT = Column(String(25, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')"))
    ReportDesc = Column(String(25, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')"))
    SAT_id = Column(UNIQUEIDENTIFIER, server_default=text("(newid())"))

    def __repr__(self):
        return str(
            self.cType.strip().title()
            + ": "
            + (
                (self.DESCRIPT.strip())
                if len(self.DESCRIPT.strip()) > len(self.ReportDesc.strip())
                else (self.ReportDesc.strip())
            )
            + " (Status ID: {})".format(str(self.SAT_id))
        )


class CSSCustStatsStore(Base):
    __tablename__ = "CSS_CustStatsStore"

    CSS_ID = Column(UNIQUEIDENTIFIER, primary_key=True, unique=True, server_default=text("(newid())"))
    CSS_Sto_pk = Column(SmallInteger, ForeignKey("SysInfo1.STO_PK"), primary_key=True, nullable=False)
    CSS_Cus_fk = Column(Integer, ForeignKey("cust.Cus_PK"), primary_key=True, nullable=False)
    CSS_TotalPawns = Column(Integer, nullable=False, server_default=text("(0)"))
    CSS_ActivePawn = Column(Integer, nullable=False, server_default=text("(0)"))
    CSS_Redeemed = Column(Integer, nullable=False, server_default=text("(0)"))
    CSS_Buys = Column(Integer, nullable=False, server_default=text("(0)"))
    CSS_Sales = Column(MONEY, nullable=False, server_default=text("(0)"))
    CSS_Amt1 = Column(MONEY, nullable=False, server_default=text("(0)"))
    CSS_Amt2 = Column(MONEY, nullable=False, server_default=text("(0)"))

    store = relationship(SysInfo1, primaryjoin=(CSS_Sto_pk == SysInfo1.STO_PK), backref="CSS_CustStatsStore")

    def __repr__(self):
        return (
                "Lifetime Pawns: {l_pawns}, "
                + "Active Pawns: {a_pawns}, "
                + "Redeemed Pawns: {r_pawns}, "
                + "Buys: {buys}, Total Sales: ${ts}, "
                + "AMT1: {amt1}, AMT2: {amt2}"
        ).format(
            l_pawns=str(self.CSS_TotalPawns),
            a_pawns=str(self.CSS_ActivePawn),
            r_pawns=str(self.CSS_Redeemed),
            buys=str(self.CSS_Buys),
            ts=str(self.CSS_Sales),
            amt1=str(self.CSS_Amt1),
            amt2=str(self.CSS_Amt2),
        )


class Cust(Base):
    __tablename__ = "cust"
    __table_args__ = (
        Index("IDX_Cust_cus_pk_cus_store", "Cus_PK", "Cus_Store"),
        Index("IDX_Cust_K1_K80", "Cus_Store", "Cus_Entered"),
        Index("IDX_Cust_K1_K2", "Cus_Store", "Cus_PK"),
        Index("IDX_cust_LName_FNAME_MNAME_IDNUM1", "CUS_LNAME", "CUS_FNAME", "CUS_MNAME", "CUS_IDNUM1"),
    )

    Cus_Store = Column(Integer, ForeignKey(SysInfo1.STO_PK), nullable=False, server_default=text("(0)"))
    Cus_PK = Column(Integer, primary_key=True, server_default=text("(0)"))
    CUS_FNAME = Column(String(15, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')"))
    CUS_MNAME = Column(String(15, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')"))
    CUS_LNAME = Column(String(25, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')"))
    CUS_ADD1 = Column(String(25, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')"))
    CUS_ADD2 = Column(String(25, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')"))
    CUS_CITY = Column(String(20, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')"))
    CUS_STATE = Column(String(2, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')"))
    CUS_ZIP = Column(String(10, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')"))
    CUS_COUNTRY = Column(String(20, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')"))
    CUS_AC1 = Column(String(3, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')"))
    CUS_PHONE1 = Column(String(8, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')"))
    CUS_HEIGHT = Column(String(6, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')"))
    CUS_WEIGHT = Column(SmallInteger, nullable=False, server_default=text("(0)"))
    CUS_HAIRFK = Column(ForeignKey("Lookup_C.lc_pk"), nullable=False, server_default=text("(0)"))
    CUS_EYESFK = Column(ForeignKey("Lookup_C.lc_pk"), nullable=False, server_default=text("(0)"))
    CUS_RACEFK = Column(ForeignKey("Lookup_C.lc_pk"), nullable=False, server_default=text("(0)"))
    CUS_SEX = Column(
        String(1, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, index=True, server_default=text("('')")
    )
    CUS_MARKS = Column(String(40, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')"))
    CUS_BIRTHDate = Column(SMALLDATETIME)
    CUS_BIRTHCITY = Column(String(15, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')"))
    CUS_BIRTHSTATE = Column(String(2, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')"))
    CUS_BIRTH2 = Column(String(15, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')"))
    CUS_IDTYP1 = Column(ForeignKey("Lookup_C.lc_pk"), nullable=False, server_default=text("(0)"))
    CUS_IDNUM1 = Column(
        String(20, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, index=True, server_default=text("('')")
    )
    CUS_ID1EXP = Column(SMALLDATETIME)
    CUS_ID1ISSUE = Column(String(50, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')"))
    CUS_IDTYP2 = Column(ForeignKey("Lookup_C.lc_pk"), nullable=False, server_default=text("(0)"))
    CUS_IDNUM2 = Column(String(20, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')"))
    CUS_ID2EXP = Column(SMALLDATETIME)
    CUS_ID2ISSUE = Column(String(2, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')"))
    CUS_KNOWN = Column(BIT, nullable=False, server_default=text("(0)"))
    CUS_SSNUM = Column(
        String(11, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, index=True, server_default=text("('   -  -    ')")
    )
    CUS_VEHIC1 = Column(String(10, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')"))
    CUS_VEHIC2 = Column(String(8, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')"))
    CUS_VEHIC3 = Column(String(2, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')"))
    CUS_EMPLOYER = Column(String(40, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')"))
    CUS_EMPAD1 = Column(String(35, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')"))
    CUS_EMPAD2 = Column(String(35, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')"))
    CUS_EMPCITY = Column(String(20, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')"))
    CUS_EMPSTATE = Column(String(2, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')"))
    CUS_EMPZIP = Column(String(10, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')"))
    CUS_AC2 = Column(String(3, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')"))
    CUS_EMPPHONE = Column(String(14, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')"))
    CUS_COMMENT = Column(String(200, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')"))
    CUS_TOTALPAWNS = Column(Integer, nullable=False, server_default=text("(0)"))
    CUS_ACTIVEPAWN = Column(Integer, nullable=False, server_default=text("(0)"))
    CUS_REDEEMED = Column(Integer, nullable=False, server_default=text("(0)"))
    CUS_BUYS = Column(Integer, nullable=False, server_default=text("(0)"))
    CUS_SALES = Column(MONEY, nullable=False, server_default=text("(0.00)"))
    CUS_AMT1 = Column(MONEY, nullable=False, server_default=text("(0.00)"))
    CUS_AMT2 = Column(MONEY, nullable=False, server_default=text("(0.00)"))
    CUS_FFLNUM = Column(String(25, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')"))
    CUS_DELETE = Column(BIT, nullable=False, server_default=text("(0)"))
    CUS_LOCKED = Column(BIT, nullable=False, server_default=text("(0)"))
    CUS_PAWNER = Column(BIT, nullable=False, server_default=text("(0)"))
    CUS_BUYER = Column(BIT, nullable=False, server_default=text("(0)"))
    CUS_IDADD1 = Column(String(25, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')"))
    CUS_IDADD2 = Column(String(25, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')"))
    CUS_IDCITY = Column(String(20, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')"))
    CUS_IDSTATE = Column(String(2, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')"))
    CUS_IDZIP = Column(String(10, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')"))
    CUS_CREDIT = Column(MONEY, nullable=False, server_default=text("(0.00)"))
    CUS_SPECIAL = Column(ForeignKey("Lookup_C.lc_pk"), nullable=False, server_default=text("(0)"))
    CUS_PIC_FK = Column(Integer, nullable=False, server_default=text("((-1))"))
    cus_Thum_fk = Column(Integer, nullable=False, server_default=text("((-1))"))
    CUS_Action_Spec = Column(CHAR(1, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("(' ')"))
    Cus_id = Column(UNIQUEIDENTIFIER, unique=True, server_default=text("(newid())"))
    Cus_PatriotAct = Column(BIT, server_default=text("(0)"))
    Cus_PointsPawnBuy = Column(Integer, nullable=False, server_default=text("(0)"))
    Cus_PointsSale = Column(Integer, nullable=False, server_default=text("(0)"))
    Cus_PointsAvail = Column(Integer, nullable=False, server_default=text("(0)"))
    Cus_PointsMailer = Column(BIT, nullable=False, server_default=text("(0)"))
    Cus_TaxID = Column(String(25, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')"))
    Cus_CellPhone = Column(String(20, "SQL_Latin1_General_CP1_CI_AS"))
    Cus_Email = Column(String(254, "SQL_Latin1_General_CP1_CI_AS"), index=True)
    Cus_Entered = Column(DateTime)
    Usr_ID = Column(UNIQUEIDENTIFIER)
    Cus_NoRemind = Column(BIT, nullable=False, server_default=text("(0)"))
    Cus_Preferred = Column(String(20, "SQL_Latin1_General_CP1_CI_AS"))
    Cus_AnnivDate = Column(DateTime)
    Cus_Search_IDNum1 = Column(String(20, "SQL_Latin1_General_CP1_CI_AS"), index=True)
    Cus_PatriotMsg = Column(BIT, server_default=text("(0)"))
    Cus_Form8300Sent = Column(DateTime)
    Cus_Search_SSNUM4 = Column(String(4, "SQL_Latin1_General_CP1_CI_AS"), index=True)
    Cus_Military = Column(BIT, server_default=text("(0)"))
    LastUpdatedUSR_ID = Column(UNIQUEIDENTIFIER)
    Cus_TxtMsg = Column(BIT, server_default=text("(0)"))
    Cus_TxtMsgSent = Column(DateTime)
    Cus_FFLExpireDate = Column(DateTime)
    Cus_TxtVerified = Column(BIT, nullable=False, server_default=text("(0)"))
    Cus_ID1IssueDate = Column(DateTime)
    Cus_ID2IssueDate = Column(DateTime)
    cus_taxexempt = Column(BIT, nullable=False, server_default=text("((0))"))
    cus_MLARate = Column(BIT, nullable=False, server_default=text("((0))"))

    Lookup_C = relationship("LookupC", primaryjoin="Cust.CUS_EYESFK == LookupC.lc_pk")
    Lookup_C1 = relationship("LookupC", primaryjoin="Cust.CUS_HAIRFK == LookupC.lc_pk")
    Lookup_C2 = relationship("LookupC", primaryjoin="Cust.CUS_IDTYP1 == LookupC.lc_pk")
    Lookup_C3 = relationship("LookupC", primaryjoin="Cust.CUS_IDTYP2 == LookupC.lc_pk")
    Lookup_C4 = relationship("LookupC", primaryjoin="Cust.CUS_RACEFK == LookupC.lc_pk")
    Lookup_C5 = relationship("LookupC", primaryjoin="Cust.CUS_SPECIAL == LookupC.lc_pk")

    stats = relationship(
        CSSCustStatsStore,
        primaryjoin=((Cus_PK == CSSCustStatsStore.CSS_Cus_fk) and (Cus_Store == CSSCustStatsStore.CSS_Sto_pk)),
        uselist=False,
        backref="Cust",
    )

    @property
    def name(self):
        return (
                self.CUS_FNAME.strip().title()
                + (" " + self.CUS_MNAME.title()[:1] if self.CUS_MNAME is not None else "")
                + " "
                + self.CUS_LNAME.strip().title()
        )

    @property
    def email(self):
        return self.Cus_Email.strip() if self.Cus_Email is not None else None

    @property
    def phone(self):
        c = (
            ("".join(x if x.isnumeric() else "" for x in self.Cus_CellPhone.strip()))
            if self.Cus_CellPhone is not None
            else ("".join(x if x.isnumeric() else "" for x in (self.CUS_AC1 + self.CUS_PHONE1)))
            if self.CUS_AC1 is not None and self.CUS_PHONE1 is not None
            else ("".join(x if x.isnumeric() else "" for x in (self.CUS_AC2 + self.CUS_PHONE2)))
            if self.CUS_AC2 is not None and self.CUS_PHONE2 is not None
            else None
        )
        return c[:10] if len(c) > 9 else None

    def __repr__(self):
        return self.name


class Acct(Base):
    __tablename__ = "Acct"

    __table_args__ = (
        Index("IDX_Acct_Sto_PK_Station", "sto_pk", "Station"),
        Index("IDX_Acct_Sto_PK_TicketNum", "TICKETNUM", "sto_pk"),
        Index("IDX_Acct_MultiKey2", "DATEin", "sto_pk", "TYPE", "TENDERTYP1", "TENDERTYP2"),
        Index("IDX_Acct_Sto_PK_Type", "sto_pk", "TYPE"),
        Index("IDX_Acct_MultiKey", "sto_pk", "TICKETNUM", "DATEin", "Acct_PK", "TYPE", "AMOUNT"),
    )

    Acct_PK = Column("Acct_PK", Integer, primary_key=True, nullable=False)
    sto_pk = Column("sto_pk", SmallInteger, nullable=False, server_default=text("(1)"))
    DATEin = Column("DATEin", DateTime, index=True)
    TICKETNUM = Column("TICKETNUM", Integer, ForeignKey("pawn.TICKETNUM"), nullable=False, server_default=text("(0)"))
    CUS_FK = Column(
        "CUS_FK", Integer, ForeignKey("cust.Cus_PK"), nullable=False, index=True, server_default=text("(0)")
    )
    TYPE = Column("TYPE", CHAR(3, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')"))
    AMOUNT = Column("AMOUNT", MONEY, nullable=False, server_default=text("(0.00)"))
    Usr_FK = Column("Usr_FK", Integer, nullable=False, server_default=text("(1)"))
    dPERCENT = Column("dPERCENT", Float(53), server_default=text("(0)"))
    TOWHOM = Column("TOWHOM", String(20, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')"))
    Enter_UsrFK = Column("Enter_UsrFK", Integer, nullable=False, server_default=text("(0)"))
    TAXSALES = Column("TAXSALES", MONEY, nullable=False, server_default=text("(0.00)"))
    STATETAX = Column("STATETAX", MONEY, nullable=False, server_default=text("(0.00)"))
    COUNTYTAX = Column("COUNTYTAX", MONEY, nullable=False, server_default=text("(0.00)"))
    LOCALTAX = Column("LOCALTAX", MONEY, nullable=False, server_default=text("(0.00)"))
    TENDERTYP1 = Column("TENDERTYP1", Integer, nullable=False, server_default=text("(0)"))
    TENDERAMT1 = Column("TENDERAMT1", MONEY, nullable=False, server_default=text("(0.00)"))
    TENDERTYP2 = Column("TENDERTYP2", Integer, nullable=False, server_default=text("(0)"))
    TENDERAMT2 = Column("TENDERAMT2", MONEY, nullable=False, server_default=text("(0.00)"))
    TENDCHANGE = Column("TENDCHANGE", MONEY, nullable=False, server_default=text("(0.00)"))
    Cus_Credit = Column("Cus_Credit", MONEY, nullable=False, server_default=text("(0.00)"))
    CreditSale = Column("CreditSale", MONEY, nullable=False, server_default=text("(0.00)"))
    Per1 = Column("Per1", MONEY, nullable=False, server_default=text("(0.00)"))
    Per2 = Column("Per2", MONEY, nullable=False, server_default=text("(0.00)"))
    OneTime = Column("OneTime", MONEY, nullable=False, server_default=text("(0.00)"))
    Ticket = Column("Ticket", MONEY, nullable=False, server_default=text("(0.00)"))
    Prep = Column("Prep", MONEY, nullable=False, server_default=text("(0.00)"))
    Gun = Column("Gun", MONEY, nullable=False, server_default=text("(0.00)"))
    Daily = Column("Daily", MONEY, nullable=False, server_default=text("(0.00)"))
    RTint = Column("RTint", MONEY, nullable=False, server_default=text("(0.00)"))
    RTsvc = Column("RTsvc", MONEY, nullable=False, server_default=text("(0.00)"))
    RTper = Column("RTper", MONEY, nullable=False, server_default=text("(0.00)"))
    RTonetime = Column("RTonetime", MONEY, nullable=False, server_default=text("(0.00)"))
    LATE = Column("LATE", MONEY, nullable=False, server_default=text("(0.00)"))
    OvrRideAmt = Column("OvrRideAmt", MONEY, nullable=False, server_default=text("(0.00)"))
    Cost = Column("Cost", MONEY, nullable=False, server_default=text("(0.00)"))
    LostTicket = Column("LostTicket", MONEY, nullable=False, server_default=text("(0.00)"))
    ExemptSales = Column("ExemptSales", MONEY, nullable=False, server_default=text("(0)"))
    cFlag = Column("cFlag", CHAR(2, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("(' ')"))
    AuctionEntry = Column("AuctionEntry", BIT, nullable=False, server_default=text("(0)"))
    RetSvcChg = Column("RetSvcChg", MONEY, nullable=False, server_default=text("(0.00)"))
    Act_id = Column("Act_id", UNIQUEIDENTIFIER, unique=True, server_default=text("(newid())"))
    Station = Column("Station", TINYINT, nullable=False, server_default=text("(1)"))
    CountyExemptSales = Column("CountyExemptSales", MONEY, nullable=False, server_default=text("(0)"))
    CountyTaxSales = Column("CountyTaxSales", MONEY, nullable=False, server_default=text("(0)"))

    customer = relationship(
        Cust, primaryjoin=((CUS_FK == Cust.Cus_PK) and (sto_pk == Cust.Cus_Store)), backref="Acct"
    )
    pawn_ticket = relationship(
        "Pawn", primaryjoin="((Acct.TICKETNUM == Pawn.TICKETNUM) and (Acct.sto_pk == Pawn.STORE_NO))", backref="Acct"
    )

    def __repr__(self):
        return str(
            "Renewal"
            if self.TYPE.strip() == "PPP"
            else "Pawn Loan"
            if self.TYPE.strip() == "P"
            else "Renewal + Principal Paydown"
            if self.TYPE.strip() == "PPL"
            else "Redemption"
            if self.TYPE.strip() == "PPU"
            else "Unknown Transaction Type ({trn})".format(trn=self.TYPE.strip())
        ) + (" ($%.2f)" % (float(self.AMOUNT) if self.TYPE.strip() != "P" else float(self.AMOUNT * -1)))


class Pawn(Base):
    __tablename__ = "pawn"
    __table_args__ = (
        Index("IDX_PAWN_MultiKey1", "STORE_NO", "usr_fk", "CUS_FK", "TICKETNUM", "DATEIN"),
        Index("IX_pawn_OrigTicket", "ORIGTICKET", "STORE_NO"),
        Index("IDX_Pawn_MultiKey3", "STATUS", "STORE_NO", "TRANS", "PWN_id"),
        Index("IDX_PAWN_MultiKey2", "TICKETNUM", "DATEIN", "STORE_NO"),
    )

    CUS_FK = Column(ForeignKey("cust.Cus_PK"), index=True, server_default=text("(1)"))
    STORE_NO = Column(SmallInteger, ForeignKey("SysInfo1.STO_PK"), nullable=False, server_default=text("(1)"))
    TICKETNUM = Column(Integer, nullable=False, server_default=text("(0)"))
    POLTICNUM = Column(Integer, nullable=False, server_default=text("(0)"))
    usr_fk = Column(Integer, nullable=False, server_default=text("(0)"))
    DATEIN = Column(SMALLDATETIME)
    NUMDAYS = Column(SmallInteger, nullable=False, server_default=text("(0)"))
    DATEOUT = Column(SMALLDATETIME, index=True)
    PawnAMT = Column(MONEY, nullable=False, server_default=text("(0.0)"))
    TRANS = Column(String(1, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('P')"))
    PAIDAMT = Column(MONEY, nullable=False, server_default=text("(0.0)"))
    TRANSDATE = Column(SMALLDATETIME)
    STATUS = Column(String(1, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')"))
    NUMITEMS = Column(SmallInteger, nullable=False, server_default=text("(0)"))
    RateTable = Column(
        String(20, "SQL_Latin1_General_CP1_CI_AS"),
        ForeignKey("c_rates.DESCRIPT"),
        nullable=False,
        server_default=text("('')"),
    )
    CHARGEDATE = Column(SMALLDATETIME)
    COMMENT = Column(String(20, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')"))
    SERVPERIOD = Column(SmallInteger, nullable=False, server_default=text("(0)"))
    COMPNDSERV = Column(BIT, nullable=False, server_default=text("(0)"))
    FLOATAMT = Column(MONEY, nullable=False, server_default=text("(0.0)"))
    GunChrg = Column(MONEY, nullable=False, server_default=text("(0.0)"))
    TICKCHRG = Column(MONEY, server_default=text("(0.0)"))
    MONTHCHRG = Column(MONEY, nullable=False, server_default=text("(0.0)"))
    MonthChrg2 = Column(MONEY, nullable=False, server_default=text("(0.0)"))
    MONTHLYCHG = Column(BIT, nullable=False, server_default=text("(0.0)"))
    PrepChrg = Column(MONEY, nullable=False, server_default=text("(0.0)"))
    StorageChrg = Column(MONEY, nullable=False, server_default=text("(0.0)"))
    REMINDER = Column(SMALLDATETIME)
    RePawned = Column(BIT, nullable=False, server_default=text("(0)"))
    ORIGTICKET = Column(Integer, nullable=False, server_default=text("(0)"))
    ORIGTICKCHRG = Column(MONEY, nullable=False, server_default=text("(0.00)"))
    ORIGONEFEE = Column(MONEY, nullable=False, server_default=text("(0)"))
    OrigDateIN = Column(SMALLDATETIME)
    OrigDateOut = Column(SMALLDATETIME)
    OrigChgDate = Column(SMALLDATETIME)
    OrigGunChrg = Column(MONEY, nullable=False, server_default=text("(0.00)"))
    OrigPawnAmt = Column(MONEY, nullable=False, server_default=text("(0.00)"))
    OrigPrepChrg = Column(MONEY, nullable=False, server_default=text("(0.00)"))
    OrigPrevPawnAmt = Column(MONEY, nullable=False, server_default=text("(0.00)"))
    OrigFloatAmt = Column(MONEY, nullable=False, server_default=text("(0.00)"))
    PrevPawnAmt = Column(MONEY, nullable=False, server_default=text("(0.00)"))
    PrevCustPK = Column(Integer, nullable=False, index=True, server_default=text("(0)"))
    PAWNNOTE = Column(String(200, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("(' ')"))
    DAILYFEE = Column(MONEY, nullable=False, server_default=text("(0.0)"))
    OvrRideAmt = Column(MONEY, nullable=False, server_default=text("(0.00)"))
    UnExtendDate = Column(SMALLDATETIME)
    StartDate = Column(SMALLDATETIME)
    TitleLoan = Column(BIT, nullable=False, server_default=text("(0)"))
    PWN_id = Column(UNIQUEIDENTIFIER, primary_key=True, unique=True, server_default=text("(newid())"))
    CheckNum = Column(String(20, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')"))
    StartPawnAmt = Column(MONEY, nullable=False, server_default=text("(0.0000)"))
    LastUpdatedUSR_ID = Column(UNIQUEIDENTIFIER)
    Text_Msg = Column(BIT, nullable=False, server_default=text("(0)"))
    PdDate = Column(SMALLDATETIME)
    OrigPdDate = Column(SMALLDATETIME)
    MLARate = Column(BIT, nullable=False, server_default=text("((0))"))

    store = relationship(
        SysInfo1, primaryjoin=(STORE_NO == SysInfo1.STO_PK), backref="pawn", uselist=False, lazy="joined"
    )

    customer = relationship(
        Cust,
        primaryjoin=(and_(foreign(CUS_FK) == Cust.Cus_PK, foreign(STORE_NO) == Cust.Cus_Store)),
        backref="pawn",
        uselist=False,
        lazy="joined",
    )

    __acct = relationship(
        Acct,
        primaryjoin=(
            and_(
                (foreign(TICKETNUM) == Acct.TICKETNUM),
                (foreign(STORE_NO) == Acct.sto_pk),
                (foreign(CUS_FK) == Acct.CUS_FK),
            )
        ),
        backref="pawn",
        uselist=True,
    )

    @property
    def pay_count(self):
        return len(
            list(filter(lambda x: x.TYPE in "PPP" and x.AMOUNT > 0 and x.sto_pk == self.STORE_NO, self.__acct))
        )

    @property
    def original_station(self):
        return (
            next(filter(lambda x: x.TYPE.strip() == "P", self.__acct)).Station
            if next(filter(lambda x: x.TYPE.strip() == "P", self.__acct), None) is not None
            else 1
        )

    _rates = relationship(
        "C_Rates",
        primaryjoin=(
            and_(
                (foreign(RateTable) == C_Rates.DESCRIPT),
                (foreign(STORE_NO) == C_Rates.STO_PK),
                (PawnAMT >= C_Rates.MINAMT),
                (PawnAMT <= C_Rates.MAXAMT),
                (PawnAMT <= C_Rates.HD),
            )
        ),
        uselist=True,
        backref="pawn",
    )

    def __repr__(self):
        return ("Pawn " if self.TRANS == "P" else "Purchase ") + (
                str(self.TICKETNUM) + " ($" + str(self.PawnAMT) + ")"
        )

    @property
    def rate(self):
        # Any given ticket's self.__rates list should only include rates whose description
        # matches, where the ticket's PawnAMT is between the rate's MIN and MAX amounts,
        # *and* whose highest dollar amount (rate.HD) is greater then or equal to the PawnAMT
        # Therefore, we should technically only have to sort out any rate whose HD amount isn't
        # the lowest in the list -and- whose PERIOD value lines up with the ticket's pay_count

        # Start by sorting out all the rates whose HD amount
        # doesn't match the lowest available amount in the rate list
        r = list(filter(lambda x: x.HD == min([x.HD for x in list(self._rates)]), list(self._rates)))

        # Check if the ticket we're dealing with has ever been paid against
        if self.pay_count == 0:
            # If it doesn't, filter out any rates whose PERIOD is greater than the lowest entry in the rate list
            r = list(filter(lambda x: x.PERIOD == min([x.PERIOD for x in r]), r))

        # If the ticket has ever been paid against,
        # check if rate list has an entry whose PERIOD matches the ticket's pay_count
        elif self.pay_count in [x.PERIOD for x in self._rates]:
            # If it does, filter out any rates that don't match
            r = list(filter(lambda x: x.PERIOD == self.pay_count, r))

        # If the ticket has a payment history and we didn't find the matching period,
        # we're probably dealing with a ticket whose payment rate is covered by the
        # entry with the highest PERIOD value
        else:
            r = list(filter(lambda x: x.PERIOD == max([x.PERIOD for x in r]), r))

        # At this point, we should have whittled the list of rates
        # down to a single correct entry, so we'll return it and call it done
        if len(r) == 1:
            return r[0]
        else:
            raise RuntimeError("Found more than one possible rate!")

    @property
    def payment_amount(self):
        return int(
            float(self.PawnAMT * ((self.rate.AP + self.rate.Interest + self.rate.MONCHRG + self.rate.ONECHRG) / 100))
            * 100
        )


class Sold(Base):
    __tablename__ = "sold"
    __table_args__ = (
        Index("IDX_Sold_Sto_PK_Sld_ID_Ticketnum", "STO_PK", "SLD_id", "TICKETNUM"),
        Index(
            "IDX_SOLD_MultiKey1", "TRANS", "STATUS", "TICKETNUM", "STO_PK", "CUS_FK", "DATEin", "DATEout", "USR_fk"
        ),
        Index("IDX_SOLD_MultiKey2", "TICKETNUM", "STO_PK", "CUS_FK", "DATEin", "TRANS"),
        Index("IDX_sold_Sto_PK_TicketNum", "TICKETNUM", "STO_PK"),
    )

    SLD_id = Column("SLD_id", UNIQUEIDENTIFIER, primary_key=True, unique=True, server_default=text("(newid())"))
    Sold_pk = Column("Sold_pk", Integer, nullable=False)
    TICKETNUM = Column("TICKETNUM", Integer, nullable=False, server_default=text("(1)"))
    STO_PK = Column(
        "STO_PK", SmallInteger, ForeignKey("SysInfo1.STO_PK"), nullable=False, index=True, server_default=text("(1)")
    )
    USR_fk = Column("USR_fk", Integer, nullable=False, server_default=text("(1)"))
    DATEin = Column("DATEin", SMALLDATETIME)
    PrevDate = Column("PrevDate", SMALLDATETIME)
    SaleAmt = Column("SaleAmt", MONEY, nullable=False, server_default=text("(0.00)"))
    Taxable = Column("Taxable", MONEY, nullable=False, server_default=text("(0.00)"))
    TAX = Column("TAX", MONEY, nullable=False, server_default=text("(0.00)"))
    ReturnedAmt = Column("ReturnedAmt", MONEY, nullable=False, server_default=text("(0.00)"))
    CUS_FK = Column(
        "CUS_FK", Integer, ForeignKey("cust.Cus_PK"), nullable=False, index=True, server_default=text("(1)")
    )
    COMMENT = Column(
        "COMMENT", String(20, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')")
    )
    TRANS = Column(
        "TRANS",
        CHAR(1, "SQL_Latin1_General_CP1_CI_AS"),
        ForeignKey("StatCode.trans"),
        nullable=False,
        server_default=text("('')"),
    )
    STATUS = Column(
        "STATUS",
        CHAR(1, "SQL_Latin1_General_CP1_CI_AS"),
        ForeignKey("StatCode.status"),
        nullable=False,
        server_default=text("('')"),
    )
    DATEout = Column("DATEout", SMALLDATETIME)
    SRVCHGPERC = Column("SRVCHGPERC", Float(24), nullable=False, server_default=text("(0.0)"))
    SRVCHGGRAC = Column("SRVCHGGRAC", SmallInteger, nullable=False, server_default=text("(0)"))
    SRVCHGAMT = Column("SRVCHGAMT", MONEY, nullable=False, server_default=text("(0.00)"))
    DEPOSIT = Column("DEPOSIT", MONEY, nullable=False, server_default=text("(0.00)"))
    PERIOD = Column("PERIOD", SmallInteger, nullable=False, server_default=text("(0)"))
    NOTE = Column("NOTE", String(40, "SQL_Latin1_General_CP1_CI_AS"), nullable=False, server_default=text("('')"))
    GunProcFee = Column("GunProcFee", MONEY, nullable=False, server_default=text("(0.00)"))
    AuctionEntry = Column("AuctionEntry", BIT, nullable=False, server_default=text("(0)"))
    CountyTaxable = Column("CountyTaxable", MONEY, nullable=False, server_default=text("(0)"))
    Reminder = Column("Reminder", SMALLDATETIME)
    sld_Message = Column("sld_Message", String(254, "SQL_Latin1_General_CP1_CI_AS"))
    LastUpdatedUSR_ID = Column("LastUpdatedUSR_ID", UNIQUEIDENTIFIER)

    store = relationship(
        SysInfo1, primaryjoin=(STO_PK == SysInfo1.STO_PK), backref="sold", uselist=False, lazy="joined"
    )

    customer = relationship(
        Cust,
        primaryjoin=(and_(foreign(CUS_FK) == Cust.Cus_PK, foreign(STO_PK) == Cust.Cus_Store)),
        backref="sold",
        uselist=False,
        lazy="joined",
    )

    status = relationship(
        StatCode,
        primaryjoin=(and_(foreign(TRANS) == StatCode.trans, foreign(STATUS) == StatCode.status)),
        backref="sold",
        uselist=False,
        lazy="joined",
    )

    def __repr__(self):
        return str(self.status.cType.strip().title() + " " + str(self.TICKETNUM))



import json
from jsonschema import validate, ValidationError
from collections import namedtuple


class SchemaValidation:
    @classmethod
    def __init__(cls):
        cls.schema_validation = namedtuple(
            typename="SchemaValidation", field_names=["valid", "error"], defaults=[False, None], module="pawn_pay"
        )

        cls.customer_upload_schema = {
            "type": "object",
            "required": ["name", "email", "phone"],
            "properties": {
                "name": {"$id": "#/properties/name", "type": "string", "examples": ["Johnny Appleseed"]},
                "email": {
                    "$id": "#/properties/email",
                    "type": "string",
                    "examples": ["j.appleseed@example.com"],
                    "pattern": "^(.*\@.*\..{3,})$",
                },
                "phone": {
                    "$id": "#/properties/phone",
                    "type": "string",
                    "examples": ["9544941112"],
                    "pattern": "^(.*)$",
                },
            },
        }

        cls.ticket_upload_schema = {
            "type": "object",
            "required": ["number", "amount", "due_date"],
            "optional": [
                "description",
                "renewal_period",
                "renew_from_payment",
                "layaway",
                "layaway_total",
                "layaway_payment",
                "customer_email",
            ],
            "properties": {
                "number": {
                    "$id": "#/properties/number",
                    "type": "string",
                    "examples": ["ZZFV-986225"],
                    "pattern": "^(.*)$",
                },
                "amount": {"$id": "#/properties/amount", "type": "integer"},
                "due_date": {
                    "$id": "#/properties/due_date",
                    "type": "date-time",
                    "examples": ["2020/07/29 03:59:59 +00:00"],
                },
                "description": {"$id": "#/properties/description", "type": "string", "default": ""},
                "renewal_period": {"$id": "#/properties/renewal_period", "type": "integer", "default": 30},
                "renew_from_payment": {
                    "$id": "#/properties/renew_from_payment",
                    "type": "string",
                    "default": "from_due_date",
                    "examples": ["from_due_date", "from_payment_date", "month", "limbo"],
                    "pattern": "^(.*)$",
                },
                "layaway": {
                    "$id": "#/properties/layaway",
                    "type": "boolean",
                    "default": False,
                    "examples": [True, False],
                },
                "layaway_total": {"$id": "#/properties/layaway_total", "type": "integer"},
                "layaway_payment": {"$id": "#/properties/layaway_payment", "type": "integer"},
                "customer_email": {
                    "$id": "#/properties/customer_email",
                    "type": "string",
                    "examples": ["shawnarnold@martinez.com"],
                    "pattern": "^(.*\@.*\..{3,})$",
                },
            },
        }

        cls.payment_upload_schema = {
            "type": "object",
            "required": ["payments"],
            "properties": {
                "payments": {
                    "$id": "#/properties/payments",
                    "type": "array",
                    "items": {
                        "$id": "#/properties/payments/items",
                        "type": "string",
                        "examples": ["ZTA3EIGL1N"],
                        "pattern": "^(.*)$",
                    },
                }
            },
        }

        cls.customer_response_schema = {
            "type": "array",
            "items": {
                "$id": "#/items",
                "type": "object",
                "required": ["id", "name", "email", "phone", "created_at", "updated_at"],
                "properties": {
                    "id": {"$id": "#/items/properties/id", "type": "integer"},
                    "name": {
                        "$id": "#/items/properties/name",
                        "type": "string",
                        "examples": ["Donald Doyle"],
                        "pattern": "^(.*)$",
                    },
                    "email": {
                        "$id": "#/items/properties/email",
                        "type": "string",
                        "examples": ["sassykatdesign@hotmail.com"],
                        "pattern": "^(.*\@.*\..{3,})$",
                    },
                    "phone": {
                        "$id": "#/items/properties/phone",
                        "type": "string",
                        "examples": ["(704) 789-7445"],
                        "pattern": "^(.*)$",
                    },
                    "created_at": {
                        "$id": "#/items/properties/created_at",
                        "type": "date-time",
                        "examples": ["2019/09/25 14:15:47 -04:00"],
                    },
                    "updated_at": {
                        "$id": "#/items/properties/updated_at",
                        "type": "date-time",
                        "examples": ["2019/09/28 11:58:28 -04:00"],
                    },
                },
            },
        }

        cls.ticket_response_schema = {
            "type": "array",
            "items": {
                "$id": "#/items",
                "type": "object",
                "required": [
                    "id",
                    "number",
                    "amount",
                    "status",
                    "due_date",
                    "description",
                    "autobill",
                    "customer_id",
                    "created_at",
                    "updated_at",
                    "scheduled_date",
                    "last_autobill",
                    "renewal_period",
                    "renew_from_payment",
                    "send_reminders",
                    "reminder_count",
                    "layaway",
                    "layaway_total",
                    "layaway_payment",
                    "payable",
                    "payment_method",
                ],
                "properties": {
                    "id": {"$id": "#/items/properties/id", "type": "integer", "default": 0, "examples": [1]},
                    "number": {
                        "$id": "#/items/properties/number",
                        "type": "string",
                        "examples": ["OUQD-645805"],
                        "pattern": "^(.*)$",
                    },
                    "amount": {
                        "$id": "#/items/properties/amount",
                        "type": "integer",
                        "default": 0,
                        "examples": [5000],
                    },
                    "status": {
                        "$id": "#/items/properties/status",
                        "type": "string",
                        "default": "active",
                        "examples": ["active", "inactive"],
                        "pattern": "^(.*)$",
                    },
                    "due_date": {
                        "$id": "#/items/properties/due_date",
                        "type": "date-time",
                        "examples": ["2019/11/25 23:59:59 -05:00"],
                    },
                    "description": {
                        "$id": "#/items/properties/description",
                        "type": "string",
                        "default": "",
                        "examples": ["Here come the future boys [future boys] - in the future, boys."],
                    },
                    "autobill": {
                        "$id": "#/items/properties/autobill",
                        "type": "boolean",
                        "default": False,
                        "examples": [True, False],
                    },
                    "customer_id": {"$id": "#/items/properties/customer_id", "type": "integer"},
                    "created_at": {
                        "$id": "#/items/properties/created_at",
                        "type": "date-time",
                        "examples": ["2019/09/25 14:37:28 -04:00"],
                    },
                    "updated_at": {
                        "$id": "#/items/properties/updated_at",
                        "type": "date-time",
                        "examples": ["2019/09/27 20:58:21 -04:00"],
                    },
                    "scheduled_date": {
                        "$id": "#/items/properties/scheduled_date",
                        "type": "None",
                        "default": None,
                        "examples": [None],
                    },
                    "last_autobill": {
                        "$id": "#/items/properties/last_autobill",
                        "type": "None",
                        "default": None,
                        "examples": [None],
                    },
                    "renewal_period": {"$id": "#/items/properties/renewal_period", "type": "integer", "default": 30},
                    "renew_from_payment": {
                        "$id": "#/items/properties/renew_from_payment",
                        "type": "string",
                        "default": "",
                        "examples": ["from_due_date"],
                        "pattern": "^(.*)$",
                    },
                    "send_reminders": {
                        "$id": "#/items/properties/send_reminders",
                        "type": "boolean",
                        "default": False,
                        "examples": [True, False],
                    },
                    "reminder_count": {"$id": "#/items/properties/reminder_count", "type": "integer", "default": 0},
                    "layaway": {
                        "$id": "#/items/properties/layaway",
                        "type": "boolean",
                        "default": False,
                        "examples": [True, False],
                    },
                    "layaway_total": {"$id": "#/items/properties/layaway_total", "type": "integer"},
                    "layaway_payment": {"$id": "#/items/properties/layaway_payment", "type": "boolean"},
                    "payable": {
                        "$id": "#/items/properties/payable",
                        "type": "boolean",
                        "default": False,
                        "examples": [True, False],
                    },
                    "payment_method": {"$id": "#/items/properties/payment_method", "type": "integer"},
                },
            },
        }

        cls.payments_recorded_schema = {
            "type": "object",
            "required": ["recorded"],
            "properties": {"recorded": {"$id": "#/properties/recorded", "type": "integer"}},
        }

        cls.new_payments_schema = {
            "type": "object",
            "required": [
                "amount",
                "created_at",
                "customer",
                "number",
                "payment_method",
                "recorded",
                "status",
                "ticket_number",
            ],
            "properties": {
                "amount": {"$id": "#/properties/amount", "type": "integer", "default": 0, "examples": [0]},
                "created_at": {
                    "$id": "#/properties/created_at",
                    "type": "date-time",
                    "examples": ["2019/04/30 08:52:17 -04:00"],
                },
                "customer": {
                    "$id": "#/properties/customer",
                    "type": "string",
                    "examples": ["jillturner@bowen.info"],
                    "pattern": "^(.*\@.*\..{3,})$",
                },
                "number": {
                    "$id": "#/properties/number",
                    "type": "string",
                    "examples": ["467529224381"],
                    "pattern": "^(.*)$",
                },
                "payment_method": {
                    "$id": "#/properties/payment_method",
                    "type": "object",
                    "required": [
                        "address",
                        "created_at",
                        "exp",
                        "expired",
                        "id",
                        "last_four",
                        "name",
                        "sub_type",
                        "type",
                        "updated_at",
                    ],
                    "properties": {
                        "address": {
                            "$id": "#/properties/payment_method/properties/address",
                            "type": "object",
                            "required": ["city", "country", "state", "street", "zip"],
                            "properties": {
                                "city": {
                                    "$id": "#/properties/payment_method/properties/address/properties/city",
                                    "type": "string",
                                    "default": "",
                                    "examples": ["East Coreneton"],
                                    "pattern": "^(.*)$",
                                },
                                "country": {
                                    "$id": "#/properties/payment_method/properties/address/properties/country",
                                    "type": "string",
                                    "default": "USA",
                                    "examples": ["USA"],
                                    "pattern": "^(.*)$",
                                },
                                "state": {
                                    "$id": "#/properties/payment_method/properties/address/properties/state",
                                    "type": "string",
                                    "default": "",
                                    "examples": ["DE"],
                                    "pattern": "^(.*)$",
                                },
                                "street": {
                                    "$id": "#/properties/payment_method/properties/address/properties/street",
                                    "type": "string",
                                    "default": "",
                                    "examples": ["91600 Jordan Forges Suite 759"],
                                    "pattern": "^(.*)$",
                                },
                                "zip": {
                                    "$id": "#/properties/payment_method/properties/address/properties/zip",
                                    "type": "string",
                                    "default": "",
                                    "examples": ["64411-8174"],
                                    "pattern": "^(.*)$",
                                },
                            },
                        },
                        "created_at": {
                            "$id": "#/properties/payment_method/properties/created_at",
                            "type": "date-time",
                            "examples": ["2019/03/26 13:27:25 -04:00"],
                        },
                        "exp": {
                            "$id": "#/properties/payment_method/properties/exp",
                            "type": "string",
                            "examples": ["1219"],
                            "pattern": "^(.*)$",
                        },
                        "expired": {
                            "$id": "#/properties/payment_method/properties/expired",
                            "type": "boolean",
                            "default": False,
                            "examples": [True, False],
                        },
                        "id": {"$id": "#/properties/payment_method/properties/id", "type": "integer"},
                        "last_four": {
                            "$id": "#/properties/payment_method/properties/last_four",
                            "type": "string",
                            "examples": ["4242"],
                            "pattern": "^(.*)$",
                        },
                        "name": {
                            "$id": "#/properties/payment_method/properties/name",
                            "type": "string",
                            "examples": ["Roger Wang"],
                            "pattern": "^(.*)$",
                        },
                        "sub_type": {
                            "$id": "#/properties/payment_method/properties/sub_type",
                            "type": "string",
                            "examples": ["visa"],
                            "pattern": "^(.*)$",
                        },
                        "type": {
                            "$id": "#/properties/payment_method/properties/type",
                            "type": "string",
                            "examples": ["credit"],
                            "pattern": "^(.*)$",
                        },
                        "updated_at": {
                            "$id": "#/properties/payment_method/properties/updated_at",
                            "type": "date-time",
                            "examples": ["2019/09/28 12:21:05 -04:00"],
                        },
                    },
                },
                "recorded": {
                    "$id": "#/properties/recorded",
                    "type": "boolean",
                    "default": False,
                    "examples": [True, False],
                },
                "status": {
                    "$id": "#/properties/status",
                    "type": "string",
                    "default": "captured",
                    "examples": ["captured", "refunded", "voided"],
                    "pattern": "^(.*)$",
                },
                "ticket_number": {
                    "$id": "#/properties/ticket_number",
                    "type": "string",
                    "examples": ["BMVQ-322975"],
                    "pattern": "^(.*)$",
                },
            },
        }

    def customer_upload(self, obj: any([dict, list])):
        try:
            return self.schema_validation(
                True if isinstance(validate(instance=obj, schema=self.customer_upload_schema), type(None)) else False,
                None,
            )
        except ValidationError as e_info:
            return self.schema_validation(False, str(e_info))

    def ticket_upload(self, obj: any([dict, list])):
        try:
            return self.schema_validation(
                True if isinstance(validate(instance=obj, schema=self.ticket_upload_schema), type(None)) else False,
                None,
            )
        except ValidationError as e_info:
            return self.schema_validation(False, str(e_info))

    def customer_response(self, obj: any([dict, list])):
        try:
            return self.schema_validation(
                True
                if isinstance(validate(instance=obj, schema=self.customer_response_schema), type(None))
                else False,
                None,
            )
        except ValidationError as e_info:
            return self.schema_validation(False, str(e_info))

    def ticket_response(self, obj: any([dict, list])):
        try:
            return self.schema_validation(
                True if isinstance(validate(instance=obj, schema=self.ticket_response_schema), type(None)) else False,
                None,
            )
        except ValidationError as e_info:
            return self.schema_validation(False, str(e_info))

    def new_payments_response(self, obj: any([dict, list])):
        try:
            return self.schema_validation(
                True if isinstance(validate(instance=obj, schema=self.new_payments_schema), type(None)) else False,
                None,
            )
        except ValidationError as e_info:
            return self.schema_validation(False, str(e_info))


def validate_json(data: str):
    """Validates that something is or can be converted to valid JSON

    Args:
        data (str): The supposedly-JSON data to validate

    Returns:
        valid (bool): True if data is JSON else False
    """
    validated = namedtuple(typename="JsonValidation", field_names=["valid", "error"], defaults=[False, None])
    try:
        json.loads(data)
        return validated(valid=True)
    except TypeError as err:
        try:
            json.loads(json.dumps(data))
            return validated(valid=True)
        except Exception as err:
            return validated(error=err)
    except json.decoder.JSONDecodeError as err:
        return validated(error=str(err))


#  Copyright (c) 2019 | Advancing Technology Systems, LLC
#  See LICENSE for any grants of usage, distribution, or modification



#  Copyright (c) 2019 | Advancing Technology Systems, LLC
#  See LICENSE for any grants of usage, distribution, or modification

from collections import namedtuple
from datetime import datetime, timedelta

from fuzzywuzzy import process
from sqlalchemy import create_engine, and_, or_
from sqlalchemy.engine.url import URL
from sqlalchemy.exc import *
from sqlalchemy.orm import sessionmaker

from pawn_pay import Customer, Ticket, Payment, CustomerList, TicketList
from pawn_pay.Exceptions import *
from .models import Cust, Pawn, LookupC, C_Rates, SysInfo1, Acct, CSSCustStatsStore

Store = namedtuple(
    typename="Store",
    field_names=["pk", "name", "address", "phone", "service_period", "pawn_days"],
    defaults=[0, "", "", "", 30, 30],
    module="PawnMaster",
)

Rate = namedtuple(
    typename="Rate",
    field_names=[
        "store_pk",
        "description",
        "multiplier",
        "period_based",
        "ap",
        "interest",
        "onetime",
        "per_month",
        "dp",
        "is_simple",
        "in_period",
        "min_amount",
        "max_amount",
    ],
    defaults=[1, "", 0.1, False, 0.0, 0.0, 0.0, 0.0, "", False, 0, 0.0, 0.0],
    module="PawnMaster",
)


class IsolinearNanoProbe(object):
    """Creates a connection to the PawnMaster DB

    NOTE:
    """

    def __init__(self, sql_driver=None, sql_user=None, sql_pass=None, sql_host=None, sql_port=None, store_pk=None):
        """Initializes and tests the connection to the SQL server

        Args:
            sql_driver (str): SQL Alchemy-supported driver to access the SQL DB
                with
            sql_user (str): Username to access SQL DB with
            sql_pass (str): Password to access SQL DB with
            sql_host (str): IP address || FQDN || DSN of SQL DB
            sql_port (str): Port SQL DB listens on
            store_pk (int): Restrict returned values to specified store

        Raises:
            CatastrophicPhaseVariance: Raised if the SQL DB cannot be accessed
                using the supplied information
        """
        self.store_pk = store_pk

        # The SQL server instance is a bit finicky about connections
        # Initialize our connection indicator to False until we can actually check that our connection is good
        self.sql_connection = False

        # Build a connection string that SQL Alchemy will recognize and use
        conn_string = URL(
            **{
                key: value
                for key, value in list(
                    filter(
                        lambda x: all([x[1] is not None, x[1] != ""]),
                        {
                            "drivername": sql_driver,
                            "username": sql_user,
                            "password": sql_pass,
                            "host": sql_host,
                            "port": sql_port,
                        }.items(),
                    )
                )
            }
        )

        self.__db = create_engine(conn_string, encoding="utf8", echo=False, implicit_returning=False)
        self.__super_session = sessionmaker(bind=self.__db)
        self.__super_session.configure(autoflush=False, autocommit=False, expire_on_commit=False)
        self.__session = self.__super_session()

        try:
            self.__db.connect()
            self.sql_connection = True
        except OperationalError as err:
            raise CatastrophicPhaseVariance(err)

        # Initialize a list to hold any errors we encounter during the
        # session, to be checked before committing changes to the DB

        self.session_errors = list()

    def __check_for_pawnbot(self):
        check = 0
        commit = False
        errors = list()

        if self.store_pk is None:
            raise CatastrophicPhaseVariance(
                "The PawnBot check cannot be executed without specifying a store_pk "
                + "during initialization of the PawnMaster connection"
            )

        if len(self.__session.execute("SELECT USR_PK FROM users WHERE USR_LANID = 'PAWNBOT';").fetchall()) < 1:
            try:
                [
                    self.__session.execute(q)
                    for q in [
                    str(
                        "INSERT INTO [dbo].[users] ([USR_PK], [USR_STORE], [USR_LANID], [USR_MENUACCESS], " +
                        "[USR_RWACCESS], [USR_ACTIVE], [USR_FNAME], [USR_STARTDATE], [PawnLimit], [BuyLimit], " +
                        "[OptionLimit], [Usr_Clock], [USR_DISCOUNT]) VALUES (161619, 1, 'PAWNBOT', 99, 99, 1, " +
                        "'PawnBot', GETDATE(), 999999.99, 999999.99, 999999.99, 1, 100);"
                    ),
                ]
                ]
                commit = True
            except Exception as err:
                errors.append(err)
        else:
            check += 1

        if len(self.__session.execute("SELECT LUC_id FROM Lookup_C WHERE lc_pk = 161619;").fetchall()) < 1:
            try:
                self.__session.execute(str("INSERT INTO [PawnMaster].[dbo].[Lookup_C] ([Sto_PK], [lc_pk], " +
                                           "[lc_Descript], [LB_FK]) VALUES (1, 161619, 'Pawn-Pay.com', 7);"))
                commit = True
            except Exception as err:
                errors.append(err)
        else:
            check += 1

        if len(self.__session.execute("SELECT ln_pk FROM Lookup_N WHERE lc_fk = 161619;").fetchall()) < 1:
            try:
                self.__session.execute(
                    str("INSERT INTO [Lookup_N] ([sto_pk], [lc_fk], [NCIC_Code], [Local_Code]) " +
                        "VALUES (1, 161619, 'Pawn-Pay.com', 'Pawn-Pay.com');")
                )
                commit = True
            except Exception as err:
                errors.append(err)
        else:
            check += 1

        if commit:
            try:
                self.__session.commit()
            except Exception as err:
                errors.append(err)

        return check, commit, errors

    @property
    def pending_changes(self):
        return {
            "to be created": self.__session.new,
            "to be updated": self.__session.dirty,
            "to be deleted": self.__session.deleted,
        }

    def save_changes(self):
        if len(self.session_errors) < 1:
            changes = {
                "new": list(self.__session.new),
                "updated": list(self.__session.dirty),
                "deleted": list(self.__session.deleted),
            }
            self.__session.flush()
            self.__session.commit()
            return True, changes
        else:
            return False, self.session_errors

    def rollback_changes(self):
        changes = {
            "not added": list(self.__session.new),
            "not updated": list(self.__session.dirty),
            "not deleted": list(self.__session.deleted),
        }
        try:
            self.__session.rollback()
            return True, changes
        except Exception as err:
            return False, err

    @property
    def store_info(self):
        info = [shop for shop in self.__session.query(SysInfo1).all()]
        return info if info.__len__() > 1 else info[0]

    def get_customers(self):
        """Get all complete customer records from PawnMaster

        Get a list of all customer records in a PawnMaster database that have
        a valid email address and phone number

        Returns:
            list: [ Customer(name (str), email (str), phone (str)), ...]
        """
        return CustomerList(
            [
                Customer(name=customer.name, email=customer.email, phone=customer.phone)
                for customer in list(
                filter(
                    lambda x: all([x.email is not None, x.phone is not None, x.name is not None]),
                    self.__session.query(Cust)
                        .filter(
                        and_(
                            Cust.Cus_Email != "",
                            Cust.Cus_CellPhone != "",
                            (
                                (Cust.Cus_Store == self.store_pk)
                                if self.store_pk is not None
                                else (Cust.Cus_Store > 0)
                            ),
                        )
                    )
                        .all(),
                )
            )
            ]
        )

    def get_tickets(self, with_customers=True):
        """Get a store's currently active ticket records from PawnMaster

        Get a list of all currently active ticket records in
        a PawnMaster database, pre-formatted for upload to Pawn-Pay.
        If no store identifier is supplied, the default store_pk (1) is used.

        Args:
            with_customers (bool): Return tickets with customer emails attached

        Returns:
            list: [ Ticket(
                        number (str),
                        amount (int),
                        due_date (datetime),
                        customer_email (str),
                        status (str),
                        renewal_period (int),
                        renew_from_payment (str),
                        layaway (bool),
                        layaway_total (int),
                        layaway_payment (int),
                    ),
                    ...
                ]
        """
        # The rate "filter" parts of the list comprehension will
        # likely need to be checked for accuracy against any
        # given PawnMaster instance before actual use

        return TicketList(
            [
                Ticket(
                    data=(
                        (
                            {
                                "number": str(t.TICKETNUM).strip(),
                                "amount": t.payment_amount,
                                "due_date": t.CHARGEDATE,
                                "customer_email": t.customer.email,
                                "renewal_period": t.store.PAWNDAYS,
                            }
                        )
                        if with_customers and t.customer is not None
                        else (
                            {
                                "number": str(t.TICKETNUM).strip(),
                                "amount": t.payment_amount,
                                "due_date": t.CHARGEDATE,
                                "renewal_period": t.store.PAWNDAYS,
                            }
                        )
                    )
                )
                for t in self.__session.query(Pawn)
                .filter(
                Pawn.STATUS == "P",
                Pawn.TRANS == "P",
                ((Pawn.STORE_NO == self.store_pk) if self.store_pk is not None else (Pawn.STORE_NO > 0)),
            )
                .all()
            ]
        )

    def update_customer(self, customer: any([dict, Customer]), id_by: str, key: any([str, int]), certainty=85):
        """Update an existing PawnMaster customer record

        Due to PawnMaster's support for multiple stores in a single DB and
        lack of enforced uniqueness for customer information, the id_by
        parameter is used to supply a strategy for ensuring the correct customer
        record is updated. Supported strategies are:

        ticket (most reliable): update the customer record associated with the supplied ticket
        email (ill advised): update any customer record(s) whose on file email address matches
        phone (ill advised): update any customer record(s) whose on file cellphone number matches
        name (don't ever do this): update any customer record(s)
                                   whose full name matches meets the minimum match certainty

        Args:
            customer (dict || pawn_pay.helpers.Customer): the new customer information
            id_by (str): the strategy used to identify the correct customer to update
            key (str): the key value used by the selected strategy to identify the correct customer to update
            certainty (int): the minimum match certainty to a customer's name required to allow an update

        Returns:
            tuple: (success (bool), customer name if successful else errors (str))
        """
        if id_by == "ticket":
            # This -should- select all tickets matching the supplied key
            # if more than one ticket is found, use fuzzywuzzy's [process.extractOne] function
            # to go through the ticket's owners to find (hopefully only) one that matches the supplied customer record

            if len("".join(str(x) if str(x).isnumeric() else "" for x in str(key))) > 1:
                lookup = [
                    x.customer
                    for x in self.__session.query(Pawn).filter(Pawn.TRANS == "P", Pawn.TICKETNUM == key).all()
                ]
                lookup = process.extractOne(customer.name, lookup) if lookup.__len__() > 0 else None
                lookup = lookup[0] if lookup is not None and lookup[1] >= certainty else None
            else:
                return False, "PawnMaster ticket numbers must be integers or numeric-only strings"

        elif id_by in ["email", "phone", "name"]:
            lookup = key.split(" ")
            lookup = (
                (self.__session.query(Cust).filter(Cust.Cus_Email == key).all())
                if id_by == "email"
                else (
                    self.__session.query(Cust)
                        .filter(
                        or_(Cust.Cus_CellPhone == key, and_(Cust.CUS_AC1 == key[:3], Cust.CUS_PHONE1 == key[3:])))
                        .all()
                )
                if id_by == "phone"
                else (
                    self.__session.query(Cust)
                        .filter(
                        Cust.CUS_FNAME.ilike("%" + lookup[0] + "%"),
                        (
                            (
                                Cust.CUS_LNAME.ilike(
                                    "%" + ((lookup[2]) if lookup.__len__() == 3 else (lookup[1])) + "%"
                                )
                            )
                            if lookup.__len__() >= 2
                            else (Cust.Cus_PK > 0)
                        ),
                    )
                        .all()
                )
            )
            lookup = process.extractOne(customer.name, lookup) if lookup.__len__() > 0 else None
            lookup = lookup[0] if lookup is not None and lookup[1] >= certainty else None

        else:
            lookup = None
            raise ValueError("Supported 'id_by' strategies are:  [ticket, email, phone, name]")

        if lookup is None:
            return False, "Could not find matching customer to update"

        try:
            up_count = 0
            if lookup.Cus_Email != customer.email:
                lookup.Cus_Email = customer.email
                up_count += 1
            if len(lookup.CUS_AC1.strip()) < 3 and len(lookup.CUS_PHONE1.strip()) < 7:
                lookup.CUS_AC1 = customer.phone[:3]
                lookup.CUS_PHONE1 = customer.phone[3:]
                up_count += 1
            if lookup.CusCellPhone != customer.phone:
                lookup.CusCellPhone = customer.phone
                up_count += 1
            if up_count > 0:
                self.__session.add(lookup)
            return True, lookup, up_count
        except Exception as err:
            self.session_errors.append(err)
            return False, err, 0

    def input_payment(self, payment: type(Payment)):
        """Input a payment made via Pawn-Pay to the PawnMaster DB

        Args:
            payment (pawn_pay.Payment): The payment to input
        """
        if len("".join(x if not x.isnumeric() else "" for x in str(payment.ticket_number))) > 0:
            return False, "Ticket numbers must be of type [int]"

        if self.store_pk is None:
            raise CatastrophicPhaseVariance(
                "Payments cannot be input without specifying a store_pk "
                + "during initialization of the PawnMaster connection"
            )

        pb_check = self.__check_for_pawnbot()

        if pb_check[0] < 3 or (pb_check[1] is True and len(pb_check[2]) > 0):
            raise CatastrophicPhaseVariance(
                "Failed existence check for the PawnBot user & tender-type!"
            )

        ticket_lookup = (
            self.__session.query(Pawn)
                .filter(
                Pawn.TICKETNUM == payment.ticket_number,
                Pawn.TRANS == "P",
                Pawn.STATUS == "P",
                Pawn.STORE_NO == self.store_pk,
            )
                .all()
        )

        if len(ticket_lookup) > 1:
            raise RuinousMicroFluctuation(
                "More than one ticket match found! Aborting attempt to process payment " + str(payment.number)
            )
        if len(ticket_lookup) < 1:
            raise RuinousMicroFluctuation(
                "No matching ticket found! Aborting attempt to process payment " + str(payment.number)
            )
        else:
            ticket_lookup = ticket_lookup[0]

        rate_total = (
                ticket_lookup.rate.AP
                + ticket_lookup.rate.Interest
                + ticket_lookup.rate.MONCHRG
                + ticket_lookup.rate.ONECHRG
        )

        int_paid = round((float(payment.amount / 100) * round(float(ticket_lookup.rate.Interest / rate_total), 5)), 2)
        svc_paid = float(payment.amount / 100) - int_paid

        payment_record = Acct(
            sto_pk=self.store_pk,
            DATEin=datetime.now(),
            TICKETNUM=payment.ticket_number,
            CUS_FK=ticket_lookup.customer.Cus_PK,
            TYPE="PPP",
            AMOUNT=float(payment.amount / 100),
            Usr_FK=161619,
            Enter_UsrFK=161619,
            TENDERTYP1=161619,
            TENDERAMT1=float(payment.amount / 100),
            OneTime=0,
            Ticket=0,
            Prep=0,
            Gun=0,
            Daily=0,
            RTint=int_paid,
            RTsvc=svc_paid,
            RTper=0,
            RTonetime=0,
            Station=ticket_lookup.original_station,
        )

        pawn_update = ticket_lookup
        pawn_update.DATEOUT = ticket_lookup.DATEOUT + timedelta(days=ticket_lookup.store.PAWNDAYS)
        pawn_update.TRANSDATE = datetime.now()
        pawn_update.CHARGEDATE = ticket_lookup.CHARGEDATE + timedelta(days=ticket_lookup.store.PAWNDAYS)
        pawn_update.PAIDAMT = float(ticket_lookup.PAIDAMT) + float(payment.amount / 100)

        customer_update = ticket_lookup.customer
        customer_update.CUS_AMT1 = float(ticket_lookup.customer.CUS_AMT1) + float(payment.amount / 100)

        customer_update.stats.CSS_Amt1 = float(customer_update.stats.CSS_Amt1) + float(payment.amount / 100)

        try:
            [self.__session.add(q) for q in [payment_record, pawn_update, customer_update]]

            return (True, payment.number) if len(self.session_errors) < 1 else (False, self.session_errors)
        except Exception as err:
            self.session_errors.append(err)
            return False, err


# Copyright (c) 2019 | Advancing Technology Systems, LLC
# See LICENSE for any grants of usage, distribution, or modification



class SeriousProblem(Exception):
    """The base Exception class that all Pawn-Pay Client & Module exceptions
    inherit from
    """
    pass


class RuinousMicroFluctuation(SeriousProblem):
    """Raised if client or module-specific behavior encounters an error

    ex. The credentials supplied to the Pawn-Pay client are invalid ex. The
    SQL Server used by PawnMaster is not reachable
    """
    pass


class CatastrophicPhaseVariance(SeriousProblem):
    """Raised if non-behavior related code encounters an error

    ex. An underlying support library throws an error that prohibits the
    functionality of the client or module, but does not crash the module itself,
    nor is it directly caused by the module or client
    """
    pass



#!/usr/local/bin/python3
#  Copyright (c) 2019 | Advancing Technology Systems, LLC
#  See LICENSE for any grants of usage, distribution, or modification

from __future__ import print_function, unicode_literals

import sentry_sdk
from sentry_sdk.integrations.sqlalchemy import SqlalchemyIntegration

sentry_sdk.init(
    dsn="https://999ed128c1f348f08be04770dd1e60b0@sentry.io/1764324", integrations=[SqlalchemyIntegration()]
)

from os import system, name
from uuid import UUID
from PyInquirer import *
from logo import print_pawn_pay_logo

import pickle
from pathlib import Path

from pprint import *


class ErrorReport(Exception):
    pass


def clear():
    _ = system("cls") if name == "nt" else system("clear")


def report_sync_error(pps_id: str, msg: str):
    raise ErrorReport(
        "Sync error reported by store [ " + pps_id + " ]. " + (("Attached message: " + msg) if msg != "" else "")
    )


def validate_uuid(uuid_string):
    try:
        val = UUID(uuid_string, version=4)
    except ValueError:
        return False
    return val.hex == uuid_string.replace("-", "").lower()


pawnpay_style = style_from_dict(
    {
        Token.Separator: "#008FFF",
        Token.QuestionMark: "#FFFB00 bold",
        Token.Selected: "#FFC100",  # default
        Token.Pointer: "#FFC100 bold",
        Token.Instruction: "",  # default
        Token.Answer: "#DDDDDD bold",
        Token.Question: "#008FFF",
    }
)

actions = [
    {
        "type": "list",
        "name": "action",
        "message": "What would you like to do?",
        "choices": [
            "[Generate] a new configuration",
            {"name": "[Edit] an existing configuration", "disabled": "Coming Soon™"},
            {"name": "[Test] an existing configuration", "disabled": "Coming Soon™"},
            Separator(),
            "[Report] a problem sync problem",
        ],
    }
]

sql_drivers = [
    "postgresql",
    "postgresql+psycopg2",
    "postgresql+pg8000",
    "postgresql+psycopg2cffi",
    "postgresql+pypostgresql",
    "postgresql+pygresql",
    "postgresql+zxjdbc",
    "mysql",
    "mysql+mysqldb",
    "mysql+pymysql",
    "mysql+mysqlconnector",
    "mysql+cymysql",
    "mysql+oursql",
    "mysql+gaerdbms",
    "mysql+pyodbc",
    "mysql+zxjdbc",
    "sqlite",
    "sqlite+pysqlite",
    "sqlite+pysqlcipher",
    "oracle",
    "oracle+cx_oracle",
    "oracle+zxjdbc",
    "mssql",
    "mssql+pyodbc",
    "mssql+mxodbc",
    "mssql+pymssql",
    "mssql+zxjdbc",
    "mssql+adodbapi",
]

prompts = [
    {
        "type": "list",
        "name": "action",
        "message": "What would you like to do?",
        "choices": [
            "[Generate] a new configuration",
            {"name": "[Edit] an existing configuration", "disabled": "Coming Soon™"},
            {"name": "[Test] an existing configuration", "disabled": "Coming Soon™"},
            Separator(),
            "[Report] a problem sync problem",
        ],
    },
    {
        "type": "input",
        "name": "pps_id",
        "message": "What's your Pawn-Pay Store ID?",
        "validate": (
            lambda val: validate_uuid(val.strip())
                        or "That doesn't look quite right... Pawn-Pay Store IDs should have some numbers, some letters, and a few dashes in them"
        ),
        "filter": lambda val: val.strip(),
        "when": lambda action: (("[Generate]" in action["action"]) or ("[Report]" in action["action"])),
    },
    {
        "type": "input",
        "name": "pps_key",
        "message": "What's your Pawn-Pay API Key?",
        "validate": lambda val: len(val.strip()) == 80
                                and not str(val.strip()).isalpha()
                                and not str(val.strip()).isnumeric()
                                and str(val.strip()).isalnum()
                                or "That doesn't look quite right...",
        "when": lambda action: "[Generate]" in action["action"],
    },
    {
        "type": "checkbox",
        "message": "What kind of records would you like to sync?",
        "name": "sync",
        "token": "> ",
        "choices": [
            Separator("-- Uncheck Record Types You Don't Want to Sync --"),
            {"name": "Tickets", "checked": True},
            {"name": "Customers", "checked": True},
            {"name": "Payments", "checked": True},
        ],
        "filter": lambda val: [x.lower() for x in val],
        "validate": (
            lambda answer: "You must choose at least *one* record type to sync" if len(answer) == 0 else True
        ),
        "when": lambda action: "[Generate]" in action["action"],
    },
    {
        "type": "list",
        "name": "data_source_type",
        "message": "Where should sync data from?",
        "choices": [
            "[A Database Server] (Used by programs like PawnMaster & PawnWizard)",
            "[A Folder -or- Files] (Used by programs like PawnDex & Hi-Tech Pawn)",
        ],
        "filter": lambda val: "path" if "Folder" in val else "sql",
        "when": lambda action: "[Generate]" in action["action"],
    },
    {
        "type": "input",
        "name": "data_path",
        "message": "Please enter the complete (absolute) path to sync data from:",
        "validate": lambda val: Path(val).exists()
                                and Path(val).is_dir()
                                or "That either doesn't appear to be a folder or it doesn't exist",
        "filter": lambda path: str(Path(path).resolve()),
        "when": lambda action: (("[Generate]" in action["action"]) and (action["data_source_type"] == "path")),
    },
    {
        "type": "input",
        "name": "driver",
        "message": "What DB driver would you like to use?",
        "default": "mssql+pymssql",
        "validate": lambda val: val in sql_drivers or "Supported drivers are: " + str(sql_drivers),
        "when": lambda action: (("[Generate]" in action["action"]) and (action["data_source_type"] == "sql")),
    },
    {
        "type": "input",
        "name": "host",
        "message": "Where is the DB you'd like to use?",
        "default": "127.0.0.1",
        "validate": lambda val: (
                                        isinstance(val, str)
                                        and not val.isnumeric()
                                        and not val.isalpha()
                                        and not val.isalnum()
                                        and val.isprintable()
        )
                                or "Please enter either an IP address or an FQDN",
        "when": lambda action: (("[Generate]" in action["action"]) and (action["data_source_type"] == "sql")),
    },
    {
        "type": "input",
        "name": "port",
        "message": "What port does the DB listen for connections on?",
        "default": "1433",
        "validate": (
            lambda val: (val.isnumeric() and not val.isalpha()) or "Please enter a number between 0 and 65535"
        ),
        "when": lambda action: (("[Generate]" in action["action"]) and (action["data_source_type"] == "sql")),
    },
    {
        "type": "input",
        "name": "db",
        "message": "Specify a DB or instance name (or leave blank):",
        "when": lambda action: (("[Generate]" in action["action"]) and (action["data_source_type"] == "sql")),
    },
    {
        "type": "list",
        "name": "trusted",
        "message": "What type of authentication should we use?",
        "choices": ["Windows Authentication", "SQL Server Authenticaton"],
        "filter": lambda val: ("SQL" not in val),
        "when": lambda action: (("[Generate]" in action["action"]) and (action["data_source_type"] == "sql")),
    },
    {
        "type": "input",
        "name": "user",
        "message": "What username should we use?",
        "default": "SA",
        "validate": lambda val: val is not None and val != "",
        "when": (
            lambda action: (
                    ("[Generate]" in action["action"])
                    and (action["data_source_type"] == "sql")
                    and (not action["trusted"])
            )
        ),
    },
    {
        "type": "password",
        "name": "password",
        "message": "What password should we use?",
        "validate": lambda val: val is not None and val != "",
        "when": (
            lambda action: (
                    ("[Generate]" in action["action"])
                    and (action["data_source_type"] == "sql")
                    and (not action["trusted"])
            )
        ),
    },
    {
        "type": "input",
        "name": "save_path",
        "message": "Where should we save the new configuration file?",
        "validate": (lambda val: Path(val).parent.resolve().exists() and not Path(val).is_dir())
                    or "That either doesn't appear to be a folder we can write to",
        "filter": lambda path: Path(path).resolve(),
        "when": lambda action: "[Generate]" in action["action"],
    },
    {
        "type": "confirm",
        "message": "Would you like to include a message?",
        "name": "include_message",
        "default": True,
        "when": lambda action: "[Report]" in action["action"],
    },
    {
        "type": "input",
        "name": "error_report",
        "message": "Please type your message and press [ENTER] when you are done:",
        "validate": lambda msg: len(msg) > 10 or "Your message should be a *least* 10 characters long",
        "when": lambda action: "[Report]" in action["action"] and action["include_message"],
    }
]

clear()
print_pawn_pay_logo("Sync Configuration Utility", "Need a hand?")

settings = prompt(prompts, style=pawnpay_style)

if "[Generate]" in settings['action']:
    save_path = settings["save_path"]

    settings = {
        "pps": {
            "id": settings["pps_id"],
            "key": settings["pps_key"],
        },
        "sync": settings["sync"],
        "data": {
            "source": settings["data_source_type"],
            "path": settings["data_path"] if "data_path" in settings.keys() else None,
            "sql": {
                "driver": settings["driver"],
                "host": settings["host"],
                "port": settings["port"],
                "user": settings["user"] if not settings["trusted"] else None,
                "password": settings["password"] if not settings["trusted"] else None,
                "trusted": settings["trusted"]
            } if settings["data_source_type"] == "sql" else None
        }
    }

    save_path.write_bytes(pickle.dumps(settings))

    print("Wrote: ")
    pprint(settings)
    print("\nTo " + str(save_path))

elif "[Report]" in settings['action']:
    try:
        report_sync_error(settings["pps_id"], msg=settings["error_report"] if settings["include_message"] else None)
    except ErrorReport as err:
        sentry_sdk.capture_exception(err)
else:
    exit()

exit()

#  Copyright (c) 2019 | Advancing Technology Systems, LLC
#  See LICENSE for any grants of usage, distribution, or modification



#  Copyright (c) 2019 | Advancing Technology Systems, LLC
#  See LICENSE for any grants of usage, distribution, or modification

from pawn_pay.PawnDex import DataFlex


class HarmonicNeutrinoField(object):
    """Interfaces with raw PawnDex files"""
    # TODO: Create class functions

    def __init__(self, pd_dir):
        """Initializes and tests the connection to the SQL server

        Args:
            pd_dir (str || Path): The directory containing PawnDex and its data files

        Raises:
            CatastrophicPhaseVariance: Raised if the supplied directory does not appear to contain a PawnDex instance
            RuinousMicroFluctuation: Raised if there is an error in handling the PawnDex files in general
        """
        pass




#  Copyright (c) 2019 | Advancing Technology Systems, LLC
#  See LICENSE for any grants of usage, distribution, or modification

from collections import namedtuple
from json import load as load_json
from pathlib import Path
from pickle import dumps as pickle_dump

from email_validator import validate_email as val, EmailNotValidError, EmailSyntaxError, EmailUndeliverableError


def validate_email(email: str):
    """Validates and formats an email address as well as checking if it is
    deliverable

    Args:
        email (str): The email to validate

    Returns:
        DubiousEmail: [valid: (bool), email: valid email || None if invalid, error (str)]
    """
    validated_email = namedtuple(
        typename="ValidatedEmail",
        field_names=["valid", "email", "error"],
        defaults=[False, None, None],
        module="pawn_pay.helpers",
    )

    try:
        return validated_email(
            True, val(email, allow_smtputf8=False, check_deliverability=True, allow_empty_local=False)["email"], None
        )
    except (EmailNotValidError, EmailSyntaxError, EmailUndeliverableError) as err:
        return validated_email(False, None, str(err))


def quick_conf(json: any([str, type(Path)]), output: any([str, type(Path)])):
    """Pickles whatever is in the supplied JSON file to quickly create
    a pawnpay client configuration without having to go all the way through
    the configuration generator

    Args:
        json (str || Path): The JSON file containing the settings to pickle
        output (str || Path): The path to output the resulting pickled settings

    Returns:
        tuple: (True if successful else False, pickle results if successful else errors)
    """
    try:
        with Path(output).resolve().open("wb") as out:
            res = out.write(pickle_dump(load_json(Path(json).resolve().open("r"))))
        return True, res
    except Exception as err:
        return False, err

# Copyright (c) 2019 | Advancing Technology Systems, LLC
# See LICENSE for any grants of usage, distribution, or modification



from setuptools import setup

setup(name='pawnpay',
      version='0.1.5',
      description='The official Python client library for the Pawn Payment Solutions API',
      keywords=['pawn', 'pay', 'pawn-pay', 'client', 'pawnmaster', 'pawndex', 'api'],
      url='https://pawn-pay.com',
      author='Pawn Payment Solutions',
      author_email='support@pawn-pay.com',
      license='GNUv3',
      packages=['pawn_pay', 'pawn_pay.helpers', 'pawn_pay.PawnMaster', 'pawn_pay.PawnDex'],
      install_requires=['Cython', 'sentry_sdk', 'loguru', 'requests', 'ratelimit', 'sqlalchemy', 'pymssql', 'pyodbc', 'fuzzywuzzy', 'python-Levenshtein-wheels'],
      zip_safe=False)


#  Copyright (c) 2019 | Advancing Technology Systems, LLC
#  See LICENSE for any grants of usage, distribution, or modification


def pawn_pay_logo(prog: str, msg: str):
    from colorama import Fore, Style
    print(Fore.BLUE + "\n          #####")
    print(Fore.BLUE + "       #    #    #           " + Fore.CYAN + "Pawn Payment Solutions")
    print(Fore.BLUE + "     #      #      #         " + Fore.WHITE + prog)
    print(Fore.BLUE + "    #       #       # ")
    print(
        Fore.YELLOW + "  #####     " + Fore.BLUE + "#     " + Fore.YELLOW + "#####      " + Fore.RED + msg
    )
    print(
        Fore.YELLOW
        + " #######    "
        + Fore.BLUE
        + "#    "
        + Fore.YELLOW
        + "#######     "
        + Fore.WHITE
        + "Email us: "
        + Fore.CYAN
        + "support@pawn-pay.com"
    )
    print(Fore.YELLOW + "  #####     " + Fore.BLUE + "#" + Fore.YELLOW + "     #####")
    print(Fore.YELLOW + "          ##### ")
    print(Fore.YELLOW + "         #######")
    print(Fore.YELLOW + "          #####\n\n")
    print(Style.RESET_ALL)



"""The Official Pawn Payment Solutions API Client"""

#  Copyright (c) 2019 | Advancing Technology Systems, LLC
#  See LICENSE for any grants of usage, distribution, or modification

import json
import logging
from collections import namedtuple, UserList
from datetime import datetime, timedelta
from pathlib import Path
from time import sleep

import requests
import sentry_sdk
from loguru import logger
from ratelimit import limits, sleep_and_retry
from sentry_sdk.integrations.logging import LoggingIntegration
from sentry_sdk.integrations.serverless import serverless_function
from sentry_sdk.integrations.sqlalchemy import SqlalchemyIntegration

from .Exceptions import *


class Customer(object):
    @classmethod
    def do_nothing(cls, key_val):
        pass

    def __init__(self, data=None, name=None, email=None, phone=None):
        if data is not None and not isinstance(data, dict):
            raise ValueError("Customer data must be supplied as a JSON array!")
        for x in [name, email, phone]:
            if x is not None and not isinstance(x, str):
                raise ValueError("Customer data must be strings when supplied via explicit keyword values!")
        if data is not None and isinstance(data, dict) and name is None and email is None and phone is None:
            super(Customer, self).__setattr__("name", data["name"])
            super(Customer, self).__setattr__("email", data["email"])
            super(Customer, self).__setattr__("phone", "".join(x if x.isnumeric() else "" for x in data["phone"]))
        elif data is None and name is not None and email is not None and phone is not None:
            super(Customer, self).__setattr__("name", name)
            super(Customer, self).__setattr__("email", email)
            super(Customer, self).__setattr__("phone", "".join(x if x.isnumeric() else "" for x in phone))
        elif data is not None and any([name is not None, email is not None, phone is not None]):
            raise AttributeError(
                "Customer objects must be instantiated with a JSON array -OR- explicit keyword values, NOT both"
            )
        elif data is None and any([name is None, email is None, phone is None]):
            raise AttributeError(
                "Customer objects must be instantiated with either a JSON array or explicit keyword values"
            )

    def __setattr__(self, key, value):
        raise NotImplementedError("Customer objects are immutable!")

    def __iter__(self):
        yield from self.__dict__.items()

    def __getitem__(self, item):
        return (
            list(self.__dict__.values())[list(self.__dict__.keys()).index(item)]
            if isinstance(item, str)
            else list(self.__dict__.values())[item]
            if isinstance(item, int)
            else ValueError
        )

    def __str__(self):
        return self.name

    def __repr__(self):
        return str(self.name + " (" + self.email + ")")


class Ticket(object):
    @classmethod
    def do_nothing(cls, key_val):
        pass

    def __init__(self, data: dict):
        from re import match

        for key, attr in data.items():
            super(Ticket, self).__setattr__(key, attr) if key in [
                "number",
                "amount",
                "status",
                "renewal_period",
                "renew_from_payment",
                "layaway",
                "layaway_total",
                "layaway_payment",
            ] else super(Ticket, self).__setattr__(
                "due_date",
                datetime.strptime(attr, "%Y/%m/%d %H:%M:%S %z")
                if isinstance(attr, str)
                else attr
                if isinstance(attr, datetime)
                else datetime.now() + timedelta(days=30),
            ) if key == "due_date" else super(
                Ticket, self
            ).__setattr__(
                "customer_email", attr
            ) if (
                (key == "customer_email") and (bool(match(r"(\w|\d|\s)*@(\w|\d|\s)*\.(\w|\d|\s){2,4}", str(attr))))
            ) else self.do_nothing(
                key_val=key
            )

    def __iter__(self):
        yield from self.__dict__.items()

    def __getitem__(self, item):
        return (
            list(self.__dict__.values())[list(self.__dict__.keys()).index(item)]
            if isinstance(item, str)
            else list(self.__dict__.values())[item]
            if isinstance(item, int)
            else ValueError
        )

    def __repr__(self):
        return str(
            "Ticket "
            + self.number
            + " (${owe:.2f}".format(owe=float(int(self.amount) / 100))
            + " due "
            + self.due_date.strftime("%m-%d-%Y")
            + ")"
        )


class Payment(object):
    @classmethod
    def do_nothing(cls, key_val):
        pass

    def __init__(self, data: dict):
        class PaymentMethod(object):
            def __init__(self, m_data: dict):
                class Address(object):
                    def __init__(self, a_data: dict):
                        [super(Address, self).__setattr__(k, v) for k, v in a_data.items()]

                    def __setattr__(self, key, value):
                        raise NotImplementedError("Address objects are immutable!")

                    def __iter__(self):
                        yield from self.__dict__.items()

                    def __getitem__(self, item):
                        return (
                            list(self.__dict__.values())[list(self.__dict__.keys()).index(item)]
                            if isinstance(item, str)
                            else list(self.__dict__.values())[item]
                            if isinstance(item, int)
                            else ValueError
                        )

                    def __str__(self):
                        return str(
                            self.street + ", " + self.city + ", " + self.state + " " + self.zip + " " + self.country
                        )

                    def __repr__(self):
                        return str(self.__dict__)

                [
                    super(PaymentMethod, self).__setattr__(
                        k,
                        (
                            v
                            if k not in ["address", "expired", "created_at", "updated_at"]
                            else Address(a_data=v)
                            if k == "address"
                            else datetime.strptime(v, "%Y/%m/%d %H:%M:%S %z")
                            if k in ["created_at", "updated_at"]
                            else v
                        ),
                    )
                    for k, v in m_data.items()
                ]

            def __setattr__(self, key, value):
                raise NotImplementedError("PaymentMethod objects are immutable!")

            def __iter__(self):
                yield from self.__dict__.items()

            def __getitem__(self, item):
                return (
                    list(self.__dict__.values())[list(self.__dict__.keys()).index(item)]
                    if isinstance(item, str)
                    else list(self.__dict__.values())[item]
                    if isinstance(item, int)
                    else ValueError
                )

            def __str__(self):
                return (
                    str(self.sub_type.title() + " Card ending in " + self.last_four)
                    if self.type == "credit"
                    else str("ACH Account ending in " + self.last_four)
                )

            def __repr__(self):
                return str({k: (v if k != "address" else self.address.__dict__) for k, v in self.__dict__.items()})

        [
            super(Payment, self).__setattr__(
                k,
                (
                    v
                    if k not in ["created_at", "payment_method", "recorded"]
                    else datetime.strptime(v, "%Y/%m/%d %H:%M:%S %z")
                    if k == "created_at"
                    else PaymentMethod(m_data=v)
                    if k == "payment_method"
                    else True
                    if k == "recorded" and v == "true"
                    else False
                    if k == "recorded" and v == "false"
                    else v
                ),
            )
            for k, v in data.items()
        ]

    def __setattr__(self, key, value):
        raise NotImplementedError("Payment objects are immutable!")

    def __iter__(self):
        yield from self.__dict__.items()

    def __getitem__(self, item):
        return (
            list(self.__dict__.values())[list(self.__dict__.keys()).index(item)]
            if isinstance(item, str)
            else list(self.__dict__.values())[item]
            if isinstance(item, int)
            else ValueError
        )

    def __repr__(self):
        return str(
            "$"
            + str(float(int(self.amount) / 100))
            + " paid against ticket "
            + str(self.ticket_number)
            + " (Ref #: "
            + self.number
            + ")"
        )


class CustomerList(UserList):
    def __init__(self, data: list):
        if isinstance(data, type(None)):
            super(CustomerList, self).__init__([])
        elif len(data) < 1:
            super(CustomerList, self).__init__([])
        else:
            super(CustomerList, self).__init__(self)
            try:
                for customer in data:
                    self.append(customer=customer)
            except KeyError:
                raise CatastrophicPhaseVariance("Customers must be passed as a list of JSON arrays")

    def append(self, customer: any([dict, Customer])):
        try:
            super(CustomerList, self).append(
                Customer(name=customer["name"], email=customer["email"], phone=customer["phone"])
                if isinstance(customer, dict)
                else customer
            )
        except KeyError:
            raise CatastrophicPhaseVariance(
                "Customers must be passed as either a <class 'pawn_pay.helpers.Customer'> "
                + 'object or a JSON array like: {"name": "Johnny Appleseed", '
                + '"email": "johnny@appleseed.com", "phone": "4045559818"}'
            )

    @property
    def for_pawnpay(self):
        cl = [x.email.lower() for x in self]

        return {
            "customers": list(
                filter(
                    lambda x: cl.count(x["email"]) == 1 and "@" in x["email"],
                    list(
                        filter(
                            lambda x: all(
                                [
                                    list(["name", "email", "phone"]) == list(x.keys()),
                                    len(list(filter(lambda y: y is None, list(x.values())))) == 0,
                                ]
                            ),
                            [x.__dict__ for x in self],
                        )
                    ),
                )
            )
        }


class TicketList(UserList):
    def __init__(self, data: list):
        if isinstance(data, type(None)):
            super(TicketList, self).__init__([])
        elif len(data) < 1:
            super(TicketList, self).__init__([])
        else:
            super(TicketList, self).__init__()
            try:
                for ticket in data:
                    self.append(ticket_data=ticket)
            except KeyError:
                raise CatastrophicPhaseVariance("Tickets must be passed as a list of JSON arrays")

    def append(self, ticket_data: any([dict, Ticket])):
        try:
            super(TicketList, self).append(Ticket(data=ticket_data) if isinstance(ticket_data, dict) else ticket_data)
        except KeyError:
            raise CatastrophicPhaseVariance(
                "Tickets must be passed as either a <class 'pawn_pay.helpers.Payment'> object or a JSON array"
            )

    @property
    def for_pawnpay(self):
        return [
            {k: v for k, v in list(filter(lambda x: all([x[1] is not None, str(x[1]) != ""]), ticket.items()))}
            for ticket in [
                {k: (v if k != "due_date" else v.strftime("%m-%d-%Y")) for k, v in item.__dict__.items()}
                for item in self
            ]
        ]


class PaymentList(UserList):
    def __init__(self, data: list):
        if isinstance(data, type(None)):
            super(PaymentList, self).__init__([])
        elif len(data) < 1:
            super(PaymentList, self).__init__([])
        else:
            super(PaymentList, self).__init__()
            try:
                for pay in data:
                    self.append(pay_data=pay)
            except KeyError:
                raise CatastrophicPhaseVariance("Payments must be passed as a list of JSON arrays")

    def append(self, pay_data: any([dict, Payment])):
        try:
            super(PaymentList, self).append(Payment(data=pay_data) if isinstance(pay_data, dict) else pay_data)
        except KeyError:
            raise CatastrophicPhaseVariance(
                "Payments must be passed as either a <class 'pawn_pay.helpers.Payment'> object or a JSON array"
            )

    @property
    def for_pawnpay(self):
        return {"payments": [x.number for x in self]}


class TemporalSubspaceHolocapacitor(object):
    """Creates a connection to the Pawn-Pay server

    The Pawn-Pay API can be broken down into three major parts - Customers, Tickets, and Payments.
    The Pawn-Pay client supplies methods for interacting with each part, and simple-to-use functions
    to perform all the actions the API allows.

    Attributes:
        authcheck (bool): A boolean indicating if the supplied Pawn-Pay API credentials are valid

    Functions:
        create_customers: Creates one or more customer records
        create_tickets: Creates one or more ticket records
        get_customers: Returns either all of a store's customers or just the specified customer
        get_tickets: Returns either all of a store's tickets or just the specified ticket
        get_new_payments: Returns all payments not marked as recorded on Pawn-Pay
        mark_payments_recorded: Marks the specified payment(s) recorded
        update_customer: Updates the name, email address, or phone number one or more customer records
        update_tickets: Updates the due date, payment amount, or active status of one or more ticket records
    """

    def __init__(self, store_id: str, api_key: str, testing=False):
        """Sets up a Pawn-Pay client instance and validates the supplied credentials

        Args:
            store_id (str): Store ID to access Pawn-Pay with
            api_key (str): API Key to access Pawn-Pay with

        Raises:
            CatastrophicPhaseVariance: Raised if the supplied Pawn-Pay
                credentials don't pass an authentication check
            RuinousMicroFluctuation: Raised if there is an error in
                communicating with Pawn-Pay in general
        """
        # First thing, setup useful logging

        # The rather odd looking if-then below ensures that the first two lines
        # of the log file contain the right strings to allow the log to be read
        # as a Markdown-formatted table instead of a traditional "log" file
        if Path("pps_sync_log.md").resolve().exists():
            with Path("pps_sync_log.md").resolve().open() as log_lines:
                log_head = [next(log_lines) for x in range(2)]
            if {"| Date / Time | Level | Message |\n", "| :---------- | :---: | :------ |\n"} != set(log_head):
                log_contents = ["| Date / Time | Level | Message |\n", "| :---------- | :---: | :------ |\n"]
                [log_contents.append(line) for line in Path("pps_sync_log.md").resolve().open().readlines()]
                with Path("pps_sync_log.md").resolve().open("w") as new_log:
                    [new_log.write(line) for line in log_contents]
                del log_contents
            del log_head
        else:
            with Path("pps_sync_log.md").resolve().open("w") as new_log:
                [
                    new_log.write(line)
                    for line in ["| Date / Time | Level | Message |\n", "| :---------- | :---: | :------ |\n"]
                ]

        self.logger = logger
        self.logger.add(
            Path("pps_sync_log.md").resolve(),
            format="{time:| YYYY-MM-DD at HH:mm:ss} | {level} | {message} | ",
            retention="60 days",
            backtrace=True,
            diagnose=True,
        )

        # Make sure that any errors are reported to Sentry, even if they occur
        # during initialization of the Pawn-Pay connection
        self.sentry = sentry_sdk

        # Technically, all of this is already happening
        # by default in sentry, but just to be pedantic...
        self.__sentry_logging = LoggingIntegration(
            level=logging.INFO,  # Capture info and above as breadcrumbs
            event_level=logging.ERROR,  # Send errors as events
        )

        self.sentry.init(
            dsn="https://999ed128c1f348f08be04770dd1e60b0@sentry.io/1764324",
            integrations=[SqlalchemyIntegration(), self.__sentry_logging],
        )

        # The client only returns named tuples
        self.__api_response = namedtuple(
            typename="ApiResponse",
            field_names=["success", "content", "status_code"],
            defaults=[False, None, None],
            module="pawn_pay",
        )

        # Set the authorization check to False until we actually check that we're authorized
        self.authcheck = False

        # Setup a connection to Pawn-Pay
        self._store_id = store_id
        self._api_key = api_key

        # Set up the request headers the way Pawn-Pay expects them to be

        self._headers = {
            r"Accept": r"application/json",
            r"Content-Type": r"application/json",
            r"User-Agent": r"pawn-pay_TemporalSubspaceHolocapacitor v0.1",
            r"X-Api-Key": self._api_key,
        }

        # The base URL for the Pawn-Pay Store
        self._baseurl = (r"https://staging." if testing else r"https://") + r"pawn-pay.com/api/v1/{storeId}".format(
            storeId=self._store_id
        )

        # Customer URIs
        self._customer_create = self._baseurl + r"/customer"  # [POST]
        self._customers_list = self._baseurl + r"/customers"  # [GET]
        self._customer = self._customer_create + r"/{email_address}"  # [GET]
        self._customers_sync = self._customers_list + r"/sync"  # [POST]

        # Ticket URIs
        self._ticket_create = self._baseurl + r"/ticket"  # [POST]
        self._tickets_batch = self._ticket_create + r"s"  # [GET] - List / [PATCH] - Update
        self._ticket = self._ticket_create + r"/{ticket}"  # [GET] - Details / [PATCH] - Update
        self._tickets_sync = self._tickets_batch + r"/sync"  # [POST]

        # Payment URIs
        self._payments = self._baseurl + r"/payments"
        self._payment = self._payments + r"/{number}"
        self._payments_new = self._payments + r"/new"  # [GET]
        self._payments_rec = self._payments + r"/recorded"  # [POST]
        self._payments_since = self._payments + r"?since={date}"  # [GET]
        self._payment_show = self._payment + r"/show"  # [GET]
        self._payment_refund = self._payment + r"/refund"  # [POST]

        # Check that the supplied credentials are viable
        try:
            auth = self.__get(url=self._baseurl)
        except (
            requests.exceptions.ConnectionError,
            requests.exceptions.ConnectTimeout,
            requests.exceptions.SSLError,
            requests.exceptions.HTTPError,
            requests.exceptions.RequestException,
        ) as err:
            raise RuinousMicroFluctuation(err)

        # Pawn-Pay returns status code '418: I am a teapot' upon successful authentication
        # but it will also return 429 for authenticated requests that have exceeded the rate limit
        # which means we can consider a 429 response status_code as a successful auth check too
        if auth.status_code in (418, 429):
            self.authcheck = True
        else:
            raise CatastrophicPhaseVariance("Authorization check failed! Bad store ID or API Key")

    @sleep_and_retry
    @limits(calls=60, period=60)  # The Pawn-Pay API limits the request rate to 60 requests per minute
    def __get(self, url: str):
        """Makes a [GET] request to the supplied URL using pre-made Pawn-Pay
        headers

        Args:
            url (str): The URL to make the [GET] request to

        Returns:
            requests.models.Response: The API response Requests object
        """
        r = requests.get(url, headers=self._headers)
        if r.status_code == 429:
            sleep(int(r.headers['Retry-After']) + 1)
        return requests.get(url, headers=self._headers) if r.status_code == 429 else r

    @sleep_and_retry
    @limits(calls=60, period=60)  # The Pawn-Pay API limits the request rate to 60 requests per minute
    def __post(self, url: str, data: dict):
        """Makes a [POST] request to the supplied URL using pre-made Pawn-Pay
        headers

        Args:
            url (str): The URL to make the [POST] request to
            data (dict): The request body to send

        Returns:
            requests.models.Response: The API response Requests object
        """
        r = requests.post(url=url, headers=self._headers, data=json.dumps(data))
        if r.status_code == 429:
            sleep(int(r.headers['Retry-After']) + 1)
        return requests.post(url=url, headers=self._headers, data=json.dumps(data)) if r.status_code == 429 else r

    @sleep_and_retry
    @limits(calls=60, period=60)  # The Pawn-Pay API limits the request rate to 60 requests per minute
    def __patch(self, url: str, data: dict):
        """Makes a [PATH] request to the supplied URL using pre-made Pawn-Pay
        headers

        Args:
            url (str): The URL to make the [PATCH] request to
            data (dict): The request body to send

        Returns:
            requests.models.Response: The API response Requests object
        """
        r = requests.patch(url=url, headers=self._headers, data=json.dumps(data))
        if r.status_code == 429:
            sleep(int(r.headers['Retry-After']) + 1)
        return requests.patch(url=url, headers=self._headers, data=json.dumps(data)) if r.status_code == 429 else r

    # noinspection DuplicatedCode
    def create_customers(self, customers: any([list, CustomerList])):
        """Create one or more customer records

        Args:
            customers (str): A list of dicts like [{"name": (str), "email": (str), "phone": (str, 10 digits)}, ...]

        Returns:
            pawn_pay.ApiResponse (tuple): [success (bool), response (JSON), status_code (int)]
        """
        data = customers.for_pawnpay if not isinstance(customers, list) else {"customers": customers}

        try:
            create = self.__post(url=self._customers_sync, data=data)
            return self.__api_response(
                True if create.status_code == 200 else False,
                create.json() if create.status_code == 200 else json.loads(create.content),
                create.status_code,
            )
        except (
            requests.exceptions.ConnectionError,
            requests.exceptions.ConnectTimeout,
            requests.exceptions.SSLError,
            requests.exceptions.HTTPError,
            requests.exceptions.RequestException,
        ) as err:
            sentry_sdk.capture_exception(err)
            return self.__api_response(False, {"errors": str(err)}, 500)

    def update_customer(self, email: str, update: any([dict, Customer])):
        """Update an individual customer record

        Args:
            email (str): The customer's email address
            update (dict): The new customer information

        Returns:
            pawn_pay.ApiResponse (tuple): [success (bool), response_body (JSON), status_code]
        """
        try:
            customer_update = self.__patch(
                url=str(self._customer.format(email_address=email)),
                data=update.__dict__ if not isinstance(update, dict) else update,
            )
            return self.__api_response(
                True if customer_update.status_code == 200 else False,
                customer_update.json() if customer_update.status_code == 200 else json.loads(customer_update.content),
                customer_update.status_code,
            )
        except (
            requests.exceptions.ConnectionError,
            requests.exceptions.ConnectTimeout,
            requests.exceptions.SSLError,
            requests.exceptions.HTTPError,
            requests.exceptions.RequestException,
        ) as err:
            sentry_sdk.capture_exception(err)
            return self.__api_response(False, {"errors": str(err)}, 500)

    # noinspection DuplicatedCode
    def create_tickets(self, data: any([list, TicketList])):
        """Creates one or more new ticket records

        Args:
            data (list): A list of dicts like:
            [
              {
                "number": (str) [omit if `ticket` is supplied,
                "amount": (int),
                "due_date": (str),
                "status": (str, [active || inactive]),
                "autobill": (bool),
                "renewal_period": (int),
                "renew_from_payment": (str)
              },
              ...
            ]

        Returns:
            pawn_pay.ApiResponse (tuple): [success (bool), response_body (JSON), status_code (int)]
        """
        created = list()
        errors = list()
        for ticket in (data.for_pawnpay) if not isinstance(data, list) else (data):
            try:
                creator = self.__post(url=self._ticket_create, data=ticket)
                created.append(ticket["number"]) if creator.status_code == 200 else errors.append(
                    tuple([ticket["number"], json.loads(creator.content)])
                )
            except (
                requests.exceptions.ConnectionError,
                requests.exceptions.ConnectTimeout,
                requests.exceptions.SSLError,
                requests.exceptions.HTTPError,
                requests.exceptions.RequestException,
            ) as err:
                sentry_sdk.capture_exception(err)
                errors.append(tuple([ticket["number"], str(err)]))
        return self.__api_response(
            True if len(errors) < 1 else False,
            {"new_count": len(created), "error_count": len(errors), "created": created, "errors": errors},
            418,
        )

    def update_tickets(self, data: any([list, Ticket, TicketList]), ticket=None):
        """Update one or more ticket records

        Args:
            data (list): A list of dicts like:
            [
              {
                "number": (str) [omit if `ticket` is supplied,
                "amount": (int),
                "due_date": (str),
                "status": (str, [active || inactive]),
                "autobill": (bool),
                "renewal_period": (int),
                "renew_from_payment": (str)
              },
              ...
            ]
            ticket (str): The ticket number to update

        Returns:
            pawn_pay.ApiResponse (tuple): [success (bool), response_body (JSON), status_code (int)]
        """
        try:
            ticket_update = self.__patch(
                url=self._ticket.format(ticket=str(data.number))
                if isinstance(data, type(Ticket))
                else self._ticket.format(ticket=ticket)
                if ticket is not None
                else self._tickets_batch,
                data=data.__dict__
                if isinstance(data, type(Ticket))
                else data.for_pawnpay
                if isinstance(data, type(TicketList))
                else data[0]
                if ticket is not None and isinstance(data, list)
                else {"tickets": data}
                if ticket is None and isinstance(data, list)
                else {},
            )
            return self.__api_response(
                True if ticket_update.status_code == 200 else False,
                ticket_update.json() if ticket_update.status_code == 200 else json.loads(ticket_update.content),
                ticket_update.status_code,
            )
        except (
            requests.exceptions.ConnectionError,
            requests.exceptions.ConnectTimeout,
            requests.exceptions.SSLError,
            requests.exceptions.HTTPError,
            requests.exceptions.RequestException,
        ) as err:
            sentry_sdk.capture_exception(err)
            return self.__api_response(False, {"errors": str(err)}, 500)

    def get_customers(self, email=None):
        """Retrieves customers from Pawn-Pay

        If `email` is not supplied returns the complete customer list, otherwise
        only the specified customer is returned

        Args:
            email (str): [Optional] Retrieve the specified customer record

        Returns:
            pawn_pay.ApiResponse (tuple): [success (bool), response (JSON), status_code (int)]
        """
        try:
            response = self.__get(
                self._customer.format(email_address=str(email.email if isinstance(email, Customer) else str(email)))
                if email is not None
                else self._customers_list
            )
            return self.__api_response(
                True if response.status_code == 200 else False,
                CustomerList(response.json())
                if response.status_code == 200 and isinstance(response.json(), list)
                else Customer(response.json())
                if response.status_code == 200 and isinstance(response.json(), dict)
                else json.loads(response.content),
                response.status_code,
            )
        except (
            requests.exceptions.ConnectionError,
            requests.exceptions.ConnectTimeout,
            requests.exceptions.SSLError,
            requests.exceptions.HTTPError,
            requests.exceptions.RequestException,
        ) as err:
            sentry_sdk.capture_exception(err)
            return self.__api_response(False, {"errors": str(err)}, 500)

    # noinspection DuplicatedCode
    def get_tickets(self, ticket=None):
        """Retrieves tickets from Pawn-Pay

        If `ticket` is not supplied returns the complete ticket list, otherwise
        only the specified ticket is returned

        Returns:
            pawn_pay.ApiResponse (tuple): [success (bool), response (JSON), status_code (int)]
        """
        try:
            tickets = self.__get(
                self._ticket.format(ticket=str(ticket.number if isinstance(ticket, Ticket) else ticket))
                if ticket is not None
                else self._tickets_batch
            )
            return self.__api_response(
                True if tickets.status_code == 200 else False,
                TicketList(tickets.json())
                if tickets.status_code == 200 and ticket is None
                else Ticket(tickets.json())
                if tickets.status_code == 200 and ticket is not None
                else json.loads(tickets.content),
                tickets.status_code,
            )
        except (
            requests.exceptions.ConnectionError,
            requests.exceptions.ConnectTimeout,
            requests.exceptions.SSLError,
            requests.exceptions.HTTPError,
            requests.exceptions.RequestException,
        ) as err:
            sentry_sdk.capture_exception(err)
            return self.__api_response(False, {"errors": str(err)}, 500)

    # noinspection DuplicatedCode
    def get_payments(self, new=False, payment=None, since=None):
        """Retrieves a list of payments that have not been marked as processed
        from Pawn-Pay

        Args:
            new (bool): Only return payments not marked as `recorded` on Pawn-Pay
            payment (str): Return on the specified payment
            since (str): Only return payments made on or after the specified date, format: MMDDYYYY

        Returns:
            pawn_pay.ApiResponse (tuple): [success (bool), response (JSON array), status_code (int)]
        """
        try:
            payments = self.__get(
                self._payments_new
                if new
                else self._payment_show.format(number=str(payment))
                if payment is not None
                else self._payments_since.format(date=str(since))
                if since is not None
                else None
            )
            return self.__api_response(
                True if payments.status_code == 200 else False,
                Payment(payments.json())
                if payments.status_code == 200 and payment is not None
                else PaymentList(payments.json())
                if payments.status_code == 200 and payment is None
                else json.loads(payments.content),
                payments.status_code,
            )
        except (
            requests.exceptions.ConnectionError,
            requests.exceptions.ConnectTimeout,
            requests.exceptions.SSLError,
            requests.exceptions.HTTPError,
            requests.exceptions.RequestException,
        ) as err:
            sentry_sdk.capture_exception(err)
            return self.__api_response(False, {"errors": str(err)}, 500)

    # noinspection DuplicatedCode
    def mark_payments_recorded(self, payments: any([str, list, PaymentList])):
        """Mark a list of payments as recorded Pawn-Pay

        Args:
            payments (list): The payments which have been recorded

        Returns:
            pawn_pay.ApiResponse (tuple): [success (bool), response (JSON), status_code (int)]
        """
        try:
            recorded_payments = self.__post(
                url=self._payments_rec,
                data=payments.for_pawnpay
                if isinstance(payments, PaymentList)
                else {"payments": payments if isinstance(payments, list) else [payments]},
            )
            return self.__api_response(
                True if recorded_payments.status_code == 200 else False,
                recorded_payments.json()
                if recorded_payments.status_code == 200
                else json.loads(recorded_payments.content),
                recorded_payments.status_code,
            )
        except (
            requests.exceptions.ConnectionError,
            requests.exceptions.ConnectTimeout,
            requests.exceptions.SSLError,
            requests.exceptions.HTTPError,
            requests.exceptions.RequestException,
        ) as err:
            sentry_sdk.capture_exception(err)
            return self.__api_response(False, {"errors": str(err)}, 500)

    # noinspection DuplicatedCode
    def refund_payment(self, payment: any([str, Payment])):
        """Mark a list of payments as recorded Pawn-Pay

        Args:
            payment (str): The payment to refund

        Returns:
            pawn_pay.ApiResponse (tuple): [success (bool), response (JSON), status_code (int)]
        """
        try:
            refund = self.__post(
                url=self._payment_refund.format(number=payment.number if isinstance(payment, Payment) else payment),
                data={"payment": payment.number if isinstance(payment, Payment) else payment},
            )
            return self.__api_response(
                True if refund.status_code == 200 else False,
                refund.json() if refund.status_code == 200 else json.loads(refund.content),
                refund.status_code,
            )
        except (
            requests.exceptions.ConnectionError,
            requests.exceptions.ConnectTimeout,
            requests.exceptions.SSLError,
            requests.exceptions.HTTPError,
            requests.exceptions.RequestException,
        ) as err:
            sentry_sdk.capture_exception(err)
            return self.__api_response(False, {"errors": str(err)}, 500)

    @serverless_function
    def sync_tickets(self, local_tickets: type(TicketList)):
        pps_tickets = TicketList(list(filter(lambda x: x.status == "active", self.get_tickets().content)))
        to_delete = TicketList(
            [
                Ticket(
                    {
                        "number": t.number,
                        "amount": t.amount,
                        "due_date": t.due_date.strftime("%Y/%m/%d %H:%M:%S %z"),
                        "status": "inactive",
                    }
                )
                for t in list(filter(lambda x: x.number not in [y.number for y in local_tickets], pps_tickets))
            ]
        )
        to_create = TicketList(
            list(
                filter(
                    lambda x: x.number not in (([y.number for y in pps_tickets]) if len(local_tickets) > 0 else ([])),
                    local_tickets,
                )
            )
        )

        to_update = TicketList([])

        for ticket in list(
            filter(
                lambda x: all(
                    [(x.number not in [y.number for y in to_delete]), (x.number not in [z.number for z in to_create])]
                ),
                local_tickets,
            )
        ):
            if (
                ticket.due_date.date()
                < next(filter(lambda x: x.number == ticket.number, pps_tickets)).due_date.date()
            ):
                if "customer_email" in ticket.__dict__.keys():
                    temp = ticket.__dict__
                    temp.__delitem__("customer_email")
                    ticket = Ticket(temp)
                to_update.append(ticket)

        return {
            "delete": ((self.update_tickets(to_delete.for_pawnpay)) if len(to_delete) > 0 else ([])),
            "create": ((self.create_tickets(to_create.for_pawnpay)) if len(to_create) > 0 else ([])),
            "update": ((self.update_tickets(to_update.for_pawnpay)) if len(to_update) > 0 else ([])),
        }


# Copyright (c) 2019 | Advancing Technology Systems, LLC
# See LICENSE for any grants of usage, distribution, or modification
